import rebound
import unittest
from spock import NbodyRegressor, DeepRegressor
import numpy as np

SAMPLE_SETTINGS = dict(samples=1000, max_model_samples=30)

def unstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-4, P=1)
    sim.add(m=1.e-4, P=1.3)
    sim.add(m=1.e-4, P=1.6)
    return sim

def longstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-7, P=1)
    sim.add(m=1.e-7, P=2.1)
    sim.add(m=1.e-7, P=4.5)
    return sim

def solarsystemsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.7e-7, a=0.39, e=0.21)
    sim.add(m=2.4e-6, a=0.72, e=0.007)
    sim.add(m=3.e-6, a=1, e=0.017)
    sim.add(m=3.2e-7, a=1.52, e=0.09)
    sim.add(m=1.e-3, a=5.2, e=0.049)
    sim.add(m=2.9e-4, a=9.54, e=0.055)
    sim.add(m=4.4e-5, a=19.2, e=0.047)
    sim.add(m=5.2e-5, a=30.1, e=0.009)
    return sim

def hyperbolicsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-5, a=-1., e=1.2)
    sim.add(m=1.e-5, a=2.)
    sim.add(m=1.e-5, a=3.)
    return sim

def escapesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-12, P=3.14, e=0.03, l=0.5)
    sim.add(m=1.e-12, P=4.396, e=0.03, l=4.8)
    sim.add(m=1.e-12, a=100, e=0.999)
    return sim

def vstablesim():
    sim = rebound.Simulation()
    sim.add(m=1)
    sim.add(m=1e-7, P=1.)
    sim.add(m=1e-7, P=1.8)
    sim.add(m=1e-7, P=3.2)
    return sim

def rescale(sim, dscale, tscale, mscale):                                                                      
    simr = rebound.Simulation()
    vscale = dscale/tscale 
    simr.G *= mscale*tscale**2/dscale**3

    for p in sim.particles:
        simr.add(m=p.m/mscale, x=p.x/dscale, y=p.y/dscale, z=p.z/dscale, vx=p.vx/vscale, vy=p.vy/vscale, vz=p.vz/vscale, r=p.r/dscale)

    return simr

class TestRegressor(unittest.TestCase):
    def setUp(self):
        self.model = DeepRegressor(cuda=False)

    def relative(self, p1, p2):
        return abs(p1 - p2) / (p1 + p2) / 2

    def test_seed(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        p1 = np.log10(self.model.predict_instability_time(sim, seed=0, **SAMPLE_SETTINGS)[0])
        p2 = np.log10(self.model.predict_instability_time(sim, seed=1, **SAMPLE_SETTINGS)[0])
        self.assertTrue(self.relative(p1, p2) < 0.1)
        self.assertTrue(self.relative(p1, p2) > 0.0)

   
    def test_prediction(self):
        times = []
        sims = []
        for mass in [1e-4, 5e-5, 3e-5, 1e-5]:
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=mass, P=1)
            sim.add(m=mass, P=1.3)
            sim.add(m=mass, P=1.6)
            sims.append(sim)

        times = np.log10(self.model.predict_instability_time(sims, **SAMPLE_SETTINGS)[0])
        # First one is unstable:
        self.assertTrue(times[0] < 4.0)
        # Should get more stable:
        self.assertTrue(np.all(times[1:] > times[:-1]))
   
    def test_rescale_distances(self):
        sim = longstablesim()
        t, upper, lower = self.model.predict_instability_time(sim, seed=0, **SAMPLE_SETTINGS)

        simr = rescale(sim, dscale=1e10, tscale=1, mscale=1)
        tr, upperr, lowerr = self.model.predict_instability_time(simr, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(t/sim.particles[1].P, tr/simr.particles[1].P, delta=np.abs((upper-lower)/10/sim.particles[1].P))
    
    def test_rescale_times(self):
        sim = longstablesim()
        t, upper, lower = self.model.predict_instability_time(sim, seed=0, **SAMPLE_SETTINGS)

        simr = rescale(sim, dscale=1, tscale=1e10, mscale=1)
        tr, upperr, lowerr = self.model.predict_instability_time(simr, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(t/sim.particles[1].P, tr/simr.particles[1].P, delta=np.abs((upper-lower)/10/sim.particles[1].P))

    def test_rescale_masses(self):
        sim = longstablesim()
        t, upper, lower = self.model.predict_instability_time(sim, seed=0, **SAMPLE_SETTINGS)

        simr = rescale(sim, dscale=1, tscale=1, mscale=1e10)
        tr, upperr, lowerr = self.model.predict_instability_time(simr, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(t/sim.particles[1].P, tr/simr.particles[1].P, delta=np.abs((upper-lower)/10/sim.particles[1].P))

    def test_time_scaling(self):
        times = []
        sims = []
        mass = 3e-5
        for P in [1, 10]:
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=mass, P=1*P)
            sim.add(m=mass, P=1.3*P)
            sim.add(m=mass, P=1.6*P)
            sims.append(sim)

        # Second time should have ~10x larger inst time.
        times = self.model.predict_instability_time(sims, seed=0, **SAMPLE_SETTINGS)[0]
        # Should be much larger time:
        self.assertGreater(times[1], 5*times[0])

    def test_time_scaling_from_integration(self):
        times = []
        sims = []
        mass = 1e-3
        for P in [1, 10]:
            sim = rebound.Simulation()
            sim.add(m=1.)
            sim.add(m=mass, P=1*P)
            sim.add(m=mass, P=1.3*P)
            sim.add(m=mass, P=1.6*P)
            sims.append(sim)

        # Second time should have ~10x larger inst time.
        times = self.model.predict_instability_time(sims, seed=0, **SAMPLE_SETTINGS)[0]
        # Should be much larger time:
        self.assertGreater(times[1], 5*times[0])

    def test_custom_prior(self):
        mass = 1e-7

        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=mass, P=1)
        sim.add(m=mass, P=1.3)
        sim.add(m=mass, P=1.6)
        expected_center = 13.0
        prior = lambda logT: np.exp(-(logT - expected_center)**2/2/0.1**2)

        times = np.log10(self.model.predict_instability_time(sim, prior_above_9=prior, **SAMPLE_SETTINGS)[0])
        self.assertAlmostEqual(times, expected_center, delta=1e-1)
    
    def test_list_time(self):
        tinst, lower, upper = self.model.predict_instability_time([hyperbolicsim(), escapesim(), unstablesim(), longstablesim()])
        self.assertTrue(np.isnan(tinst[0]))
        self.assertLess(tinst[1], 1e4)
        self.assertLess(tinst[2], 1e4)
        self.assertGreater(tinst[3], 1e4)

class TestRegressorClassification(unittest.TestCase):
    def setUp(self):
        self.model = DeepRegressor(cuda=False)
    
    def test_list_stable(self): # pass list of sims with same size list of tmax
        tmax = [1e4, 1e4, 1, 1e4] # test that unstablesim in middle still classified as stable with tmax=1
        stable_target = [0, 0, 1, 1]
        stable = self.model.predict_stable([hyperbolicsim(), escapesim(), unstablesim(), longstablesim()], tmax=tmax)
        self.assertSequenceEqual(stable.tolist(), stable_target)
    
    def test_list_no_tmax(self): # pass list of sims, tmax = None
        stable = self.model.predict_stable([vstablesim(), vstablesim()])
        self.assertGreater(stable[0], 0.9)
        self.assertGreater(stable[1], 0.9)
    
    def test_single_no_tmax(self): # pass list of sims, tmax = None
        stable = self.model.predict_stable(vstablesim())
        self.assertGreater(stable, 0.9)
    
    def test_mismatched_lists(self):
        tmax = [1e4, 1e4, 1] # test that unstablesim in middle still classified as stable with tmax=1
        with self.assertRaises(AssertionError):
            stable = self.model.predict_stable([hyperbolicsim(), escapesim(), unstablesim(), longstablesim()], tmax=tmax)

    def test_sim_unchanged(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        sim.integrate(1.2)
        x0 = sim.particles[1].x
        p1 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertEqual(sim.particles[1].x, x0)

    def test_repeat(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        p1 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        p2 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertEqual(p1, p2)
   
    # when chaotic realization matters, probs will vary more (eg t_inst=2e4)
    def test_galilean_transformation(self):
        sim = longstablesim()
        sim.move_to_com()
        p_com = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)

        sim = longstablesim()
        for p in sim.particles:
            p.vx += 1000
        p_moving = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(p_com, p_moving, delta=1e-2)
   
    def test_rescale_distances(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)

        sim = longstablesim()
        sim = rescale(sim, dscale=1e10, tscale=1, mscale=1)
        p1 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(p0, p1, delta=1e-2)
    
    def test_rescale_times(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)

        sim = longstablesim()
        sim = rescale(sim, dscale=1, tscale=1e10, mscale=1)
        p1 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(p0, p1, delta=1e-2)

    def test_rescale_masses(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)

        sim = longstablesim()
        sim = rescale(sim, dscale=1, tscale=1, mscale=1e10)
        p1 = self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS)
        self.assertAlmostEqual(p0, p1, delta=1e-2)
    
    def test_hyperbolic(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, a=-1., e=1.2)
        sim.add(m=1.e-5, a=2.)
        sim.add(m=1.e-5, a=3.)
        self.assertEqual(self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS), 0)
    
    def test_escape(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-12, P=3.14, e=0.03, l=0.5)
        sim.add(m=1.e-12, P=4.396, e=0.03, l=4.8)
        sim.add(m=1.e-12, a=100, e=0.999)
        self.assertEqual(self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS), 0)
    
    def test_unstable_in_short_integration(self):
        sim = unstablesim()
        self.assertEqual(self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS), 0)
    
    def test_solarsystem(self):
        sim = solarsystemsim()
        median, lower, upper, t_inst_samples = self.model.predict_instability_time(
                sim, seed=0,
                return_samples=True,
                **SAMPLE_SETTINGS)
        log_iqr = np.log10(upper) - np.log10(lower)
        correct_estimate = (np.average(t_inst_samples > 1e9) > 0.7 )
        very_uncertain = (log_iqr > 3)
        self.assertTrue(correct_estimate or very_uncertain)

    def test_stable(self):
        sim = longstablesim()
        self.assertGreater(self.model.predict_stable(sim, seed=0, **SAMPLE_SETTINGS), 0.7)

if __name__ == '__main__':
    unittest.main()



import pandas as pd
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_curve, confusion_matrix, auc
from sklearn import metrics
import numpy as np

def hasnull(row):
    numnulls = row.isnull().sum()
    if numnulls == 0:
        return 0
    else:
        return 1

def train_test_split(trainingdatafolder, features=None, labelname='Stable', filter=False, filtertimes=False):
    dataset = pd.read_csv(trainingdatafolder+"trainingdata.csv", index_col = 0)
    if features is None:
        features = dataset.columns.values
    dataset['hasnull'] = dataset.apply(hasnull, axis=1)

    labels = pd.read_csv(trainingdatafolder+"labels.csv", index_col=0)
    if filter:
        y = labels[(labels['instability_time'] > 1.e4) & (dataset['hasnull'] == 0)][labelname]
        X = dataset[(labels['instability_time'] > 1.e4) & (dataset['hasnull'] == 0)][features]
    elif filtertimes:
        y = labels[labels['instability_time'] > 1.e4][labelname]
        X = dataset[labels['instability_time'] > 1.e4][features]
    else:
        y = labels[labelname]
        X = dataset[features]

    Nrows = int(0.8*X.shape[0])
    trainX = X.iloc[:Nrows, :]
    trainy = y.iloc[:Nrows]
    testX = X.iloc[Nrows:, :]
    testy = y.iloc[Nrows:]

    return trainX, trainy, testX, testy

def ROC_curve(trainingdatafolder, model, features=None, filter=False, filtertimes=False):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]
    fpr, tpr, ROCthresholds = roc_curve(testy, preds)
    roc_auc = metrics.roc_auc_score(testy, preds)
    return roc_auc, fpr, tpr, ROCthresholds

def PR_curve(trainingdatafolder, model, features=None, filter=False, filtertimes=False):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]
    precision, recall, PRthresholds = precision_recall_curve(testy, preds)
    pr_auc = metrics.auc(recall, precision)
    return pr_auc, precision, recall, PRthresholds

def tnr_npv_curve(trainingdatafolder, model, features=None, filter=False, filtertimes=False, N=1000):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]
    npv, tnr = np.zeros(N), np.zeros(N)
    thresholds = np.linspace(0, 1, N)
    for i, thresh in enumerate(thresholds):
        predy = (preds >= thresh)
        tn, fp, fn, tp = confusion_matrix(testy, predy).ravel()
        npv[i] = tn/(tn+fn)
        tnr[i] = tn/(tn+fp)
    # When thresh is 0, we don't predict any negs, so npv is nan
    npv[np.isnan(npv)] = 1
    aucval = auc(tnr, npv)
    return aucval, npv, tnr, thresholds

def stable_unstable_hist(trainingdatafolder, model, features=None, filter=False, filtertimes=False):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]
    stablepreds = preds[np.where(testy==1)]
    unstablepreds = preds[np.where(testy==0)]
    return stablepreds, unstablepreds 

def calibration_plot(trainingdatafolder, model, features=None, bins=10, filter=False, filtertimes=False):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]

    hist, edges = np.histogram(preds, bins=bins)

    bincenters = []
    fracstable = []
    errorbars = []
    for i in range(len(edges)-1):
        bincenters.append((edges[i]+edges[i+1])/2)
        mask = (preds >= edges[i]) & (preds < edges[i+1])
        nstable = testy[mask].sum()
        fracstable.append(nstable/hist[i]) # fraction of stable systems in bin with predictions in range
        errorbars.append(np.sqrt(1./nstable + 1./hist[i])*fracstable[-1]) # assume poisson counting errors for each fractional error, and add in quadrature for error on ratio. 
        # multiply the fractional error by value

    return np.array(bincenters), np.array(fracstable), np.array(errorbars)

def unstable_error_fraction(trainingdatafolder, model, thresh, features=None, bins=10, filter=False, filtertimes=False):
    trainX, trainy, testX, testy = train_test_split(trainingdatafolder, features, filter=filter, filtertimes=filtertimes)
    preds = model.predict_proba(testX)[:,1]
    dummy, dummy, dummy, inst_times = train_test_split(trainingdatafolder, features, labelname='instability_time', filter=filter, filtertimes=filtertimes)
    log_inst_times = np.log10(inst_times)
    
    unstable = log_inst_times < 8.99
    preds = preds[unstable]
    log_inst_times = log_inst_times[unstable]

    hist, edges = np.histogram(log_inst_times, bins=bins)

    bincenters = []
    errorfracs = []
    errorbars = []
    for i in range(len(edges)-1):
        bincenters.append((edges[i]+edges[i+1])/2)
        mask = (log_inst_times >= edges[i]) & (log_inst_times < edges[i+1])
        Nerrors = (preds[mask] > thresh).sum()
        errorfracs.append(Nerrors/hist[i])
        errorbars.append(np.sqrt(1./Nerrors + 1./hist[i])*errorfracs[-1]) # see calibration plot comment

    return np.array(bincenters), np.array(errorfracs), np.array(errorbars)



import numpy as np
import rebound

def check_hyperbolic(sim):
    orbits = sim.calculate_orbits()
    amin = np.min([o.a for o in orbits])
    if amin < 0: # at least one orbit is hyperbolic (a<0)
        return True
    else:
        return False

def check_valid_sim(sim):
    assert isinstance(sim, rebound.Simulation)
    ps = sim.particles
    ms = np.array([p.m for p in sim.particles[:sim.N_real]])
    if np.min(ms) < 0: # at least one body has a mass < 0
        raise AttributeError("SPOCK Error: Particles in sim passed to spock_features had negative masses")

    if np.max(ms) != ms[0]:
        raise AttributeError("SPOCK Error: Particle at index 0 must be the primary (dominant mass)")

    return

def set_integrator_and_timestep(sim):
    Ps = np.array([p.P for p in sim.particles[1:sim.N_real]])
    es = np.array([p.e for p in sim.particles[1:sim.N_real]])
    if np.max(es) < 1:
        minTperi = np.min(Ps*(1-es)**1.5/np.sqrt(1+es)) # min peri passage time
        sim.dt = 0.05*minTperi                          # Wisdom 2015 suggests 0.05
    else:                                               # hyperbolic orbit 
        sim.dt = np.nan # so tseries gives nans, but still always gives same shape array

    if np.max(es) > 0.99:                               # avoid stall with WHFAST for e~1
        sim.integrator = "ias15"
    else:
        sim.integrator = "whfast"

def init_sim_parameters(sim, megno=True, safe_mode=1): 
    # if megno=False and safe_mode=0, integration will be 2x faster. But means we won't get the same trajectory realization for the systems in the training set, but rather a different equally valid realization. We've tested that this doesn't affect the performance of the model (as it shouldn't!).

    check_valid_sim(sim)

    try:
        sim.collision = 'line'  # use line if using newer version of REBOUND
    except:
        sim.collision = 'direct'# fall back for older versions

    maxd = np.array([p.d for p in sim.particles[1:sim.N_real]]).max()
    sim.exit_max_distance = 100*maxd
                
    sim.ri_whfast.keep_unsynchronized = 0
    sim.ri_whfast.safe_mode = safe_mode 

    if sim.N_var == 0 and megno: # no variational particles
        sim.init_megno(seed=0)
   
    set_integrator_and_timestep(sim)

    # Set particle radii to their individual Hill radii. 
    # Exact collision condition doesn't matter, but this behaves at extremes.
    # Imagine huge M1, tiny M2 and M3. Don't want to set middle planet's Hill 
    # sphere to mutual hill radius with huge M1 when catching collisions w/ M3
    
    for p in sim.particles[1:sim.N_real]:
        rH = p.a*(p.m/3./sim.particles[0].m)**(1./3.)
        p.r = rH
    
    sim.move_to_com()



from runfunctions import run_random
import sys

sim_id = int(sys.argv[1])
runstring = sys.argv[2]
run_random(sim_id, runstr=runstring) # run system
run_random(sim_id, shadow=True, runstr=runstring) # run shadow system



import rebound
import numpy as np
from collections import OrderedDict

######################### Taken from celmech github.com/shadden/celmech
def farey_sequence(n):
    """Return the nth Farey sequence as order pairs of the form (N,D) where `N' is the numerator and `D' is the denominator."""
    a, b, c, d = 0, 1, 1, n
    sequence=[(a,b)]
    while (c <= n):
        k = int((n + b) / d)
        a, b, c, d = c, d, (k*c-a), (k*d-b)
        sequence.append( (a,b) )
    return sequence
def resonant_period_ratios(min_per_ratio,max_per_ratio,order):
    """Return the period ratios of all resonances up to order 'order' between 'min_per_ratio' and 'max_per_ratio' """
    if min_per_ratio < 0.:
        raise AttributeError("min_per_ratio of {0} passed to resonant_period_ratios can't be < 0".format(min_per_ratio))
    if max_per_ratio >= 1.:
        raise AttributeError("max_per_ratio of {0} passed to resonant_period_ratios can't be >= 1".format(max_per_ratio))
    minJ = int(np.floor(1. /(1. - min_per_ratio)))
    maxJ = int(np.ceil(1. /(1. - max_per_ratio)))
    res_ratios=[(minJ-1,minJ)]
    for j in range(minJ,maxJ):
        res_ratios = res_ratios + [ ( x[1] * j - x[1] + x[0] , x[1] * j + x[0]) for x in farey_sequence(order)[1:] ]
    res_ratios = np.array(res_ratios)
    msk = np.array( list(map( lambda x: min_per_ratio < x[0]/float(x[1]) < max_per_ratio , res_ratios )) )
    return res_ratios[msk]
##########################

# sorts out which pair of planets has a smaller EMcross, labels that pair inner, other adjacent pair outer
# returns a list of two lists, with [label (near or far), i1, i2], where i1 and i2 are the indices, with i1 
# having the smaller semimajor axis
def get_pairs(sim, indices):
    ps = sim.particles
    sortedindices = sorted(indices, key=lambda i: ps[i].a) # sort from inner to outer
    EMcrossInner = (ps[sortedindices[1]].a-ps[sortedindices[0]].a)/ps[sortedindices[0]].a
    EMcrossOuter = (ps[sortedindices[2]].a-ps[sortedindices[1]].a)/ps[sortedindices[1]].a

    if EMcrossInner < EMcrossOuter:
        return [['near', sortedindices[0], sortedindices[1]], ['far', sortedindices[1], sortedindices[2]]]
    else:
        return [['near', sortedindices[1], sortedindices[2]], ['far', sortedindices[0], sortedindices[1]]]

def find_strongest_MMR(sim, i1, i2):
    maxorder = 2
    ps = sim.particles
    n1 = ps[i1].n
    n2 = ps[i2].n

    m1 = ps[i1].m/ps[0].m
    m2 = ps[i2].m/ps[0].m

    Pratio = n2/n1

    delta = 0.03
    if Pratio < 0 or Pratio > 1: # n < 0 = hyperbolic orbit, Pratio > 1 = orbits are crossing
        return np.nan, np.nan, np.nan

    minperiodratio = max(Pratio-delta, 0.)
    maxperiodratio = min(Pratio+delta, 0.99) # too many resonances close to 1
    res = resonant_period_ratios(minperiodratio,maxperiodratio, order=2)

    # Calculating EM exactly would have to be done in celmech for each j/k res below, and would slow things down. This is good enough for approx expression
    EM = np.sqrt((ps[i1].e*np.cos(ps[i1].pomega) - ps[i2].e*np.cos(ps[i2].pomega))**2 + (ps[i1].e*np.sin(ps[i1].pomega) - ps[i2].e*np.sin(ps[i2].pomega))**2)
    EMcross = (ps[i2].a-ps[i1].a)/ps[i1].a

    j, k, maxstrength = np.nan, np.nan, 0 
    for a, b in res:
        nres = (b*n2 - a*n1)/n1
        if nres == 0:
            s = np.inf # still want to identify as strongest MMR if initial condition is exatly b*n2-a*n1 = 0
        else:
            s = np.abs(np.sqrt(m1+m2)*(EM/EMcross)**((b-a)/2.)/nres)
        if s > maxstrength:
            j = b
            k = b-a
            maxstrength = s
    if maxstrength == 0:
        maxstrength = np.nan

    return j, k, maxstrength

def populate_trio(sim, trio, pairs, tseries, i):
    Ns = 3
    ps = sim.particles
    for q, [label, i1, i2] in enumerate(pairs):
        m1 = ps[i1].m
        m2 = ps[i2].m
        e1x, e1y = ps[i1].e*np.cos(ps[i1].pomega), ps[i1].e*np.sin(ps[i1].pomega)
        e2x, e2y = ps[i2].e*np.cos(ps[i2].pomega), ps[i2].e*np.sin(ps[i2].pomega)
        tseries[i,Ns*q+1] = np.sqrt((e2x-e1x)**2 + (e2y-e1y)**2)
        tseries[i,Ns*q+2] = np.sqrt((m1*e1x + m2*e2x)**2 + (m1*e1y + m2*e2y)**2)/(m1+m2)
        j, k, tseries[i,Ns*q+3] = find_strongest_MMR(sim, i1, i2) 

    tseries[i,7] = sim.calculate_megno() # megno

def get_tseries(sim, args):
    Norbits = args[0]
    Nout = args[1]
    trios = args[2]
    
    minP = np.min([p.P for p in sim.particles[1:sim.N_real]])

    # want hyperbolic case to run so it raises exception
    times = np.linspace(0, Norbits*np.abs(minP), Nout)
    
    triopairs, triotseries = [], []
    for tr, trio in enumerate(trios): # For each trio there are two adjacent pairs 
        triopairs.append(get_pairs(sim, trio))
        triotseries.append(np.zeros((Nout, 8))*np.nan)

    if np.isnan(sim.dt): # initially hyperbolic orbit, set in simsetup/set_integrator_and_timestep
        stable = False
        return triotseries, stable

    for i, time in enumerate(times):
        try:
            sim.integrate(time, exact_finish_time=0)
        except (rebound.Collision, rebound.Escape):
            stable = False
            return triotseries, stable

        for tseries in triotseries:
            tseries[i,0] = sim.t/minP  # time

        for tr, trio in enumerate(trios):
            pairs = triopairs[tr]
            tseries = triotseries[tr] 
            populate_trio(sim, trio, pairs, tseries, i)
    
    stable = True
    return triotseries, stable
    
def features(sim, args):
    Norbits = args[0]
    Nout = args[1]
    trios = args[2]
    
    ps  = sim.particles
    triofeatures = []
    for tr, trio in enumerate(trios):
        features = OrderedDict()
        pairs = get_pairs(sim, trio)
        for i, [label, i1, i2] in enumerate(pairs):
            features['EMcross'+label] = (ps[i2].a-ps[i1].a)/ps[i1].a
            features['EMfracstd'+label] = np.nan
            features['EPstd'+label] = np.nan
            features['MMRstrength'+label] = np.nan

        features['MEGNO'] = np.nan
        features['MEGNOstd'] = np.nan
        triofeatures.append(features)
    
    triotseries, stable = get_tseries(sim, args)
    if stable == False:
        return triofeatures, stable

    for features, tseries in zip(triofeatures, triotseries):
        EMnear = tseries[:, 1]
        EPnear = tseries[:, 2]
        # cut out first value (init cond) to avoid cases
        # where user sets exactly b*n2 - a*n1 & strength is inf
        MMRstrengthnear = tseries[1:,3]
        EMfar = tseries[:, 4]
        EPfar = tseries[:, 5]
        MMRstrengthfar = tseries[1:,6]
        MEGNO = tseries[:, 7]

        if not np.isnan(MEGNO).any(): # no nans
            features['MEGNO'] = np.median(MEGNO[-int(Nout/10):]) # smooth last 10% to remove oscillations around 2
            features['MEGNOstd'] = MEGNO[int(Nout/5):].std()
        features['MMRstrengthnear'] = np.median(MMRstrengthnear)
        features['MMRstrengthfar'] = np.median(MMRstrengthfar)
        features['EMfracstdnear'] = EMnear.std() / features['EMcrossnear']
        features['EMfracstdfar'] = EMfar.std() / features['EMcrossfar']
        features['EPstdnear'] = EPnear.std() 
        features['EPstdfar'] = EPfar.std() 
    
    return triofeatures, stable



import numpy as np
import math
from scipy.stats import truncnorm
import os
from collections import OrderedDict
from .tseries_feature_functions import get_extended_tseries
from copy import deepcopy as copy
import torch
from torch import nn
from torch.nn import Parameter
from torch.autograd import Variable
from torch.functional import F
import glob
from .spock_reg_model import load_swag
from pytorch_lightning import Trainer
import torch
import time
import pickle as pkl
import warnings
import einops as E
from scipy.integrate import quad
from scipy.interpolate import interp1d
import pytorch_lightning as pl
from .simsetup import init_sim_parameters
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool as Pool
import rebound
import random
warnings.filterwarnings('ignore', "DeprecationWarning: Using or importing the ABCs")

def fitted_prior():
    """A prior that was fit to the PDF of instability times in the training set below T=9"""
    return (lambda logT: 3.27086190404742*np.exp(-0.424033970670719 * logT) - 10.8793430454878*np.exp(-0.200351029031774 * logT**2))

def flat_prior(upper_limit):
    """A uniform prior between 9 and an upper limit"""
    return (lambda logT: 1.0 * (logT <= upper_limit))

def exponential_decaying_prior(decay_rate):
    """An exponentially decaying prior, which decays at a rate e^(-T*decay_rate)"""
    return (lambda logT: np.exp(-decay_rate * (logT - 9)))

profile = lambda _: _

def generate_dataset(sim): 
    sim = sim.copy()
    init_sim_parameters(sim, megno=False, safe_mode=0)
    if sim.N_real < 4:
        raise AttributeError("SPOCK Error: SPOCK only works for systems with 3 or more planets") 
    trios = [[i,i+1,i+2] for i in range(1,sim.N_real-2)] # list of adjacent trios

    kwargs = OrderedDict()
    kwargs['Norbits'] = int(1e4)
    kwargs['Nout'] = 100
    kwargs['trios'] = trios
    args = list(kwargs.values())
    # These are the .npy.
    # In the other file, we concatenate (restseries, orbtseries, mass_array)
    tseries, stable = get_extended_tseries(sim, args, mmr=False, megno=False)

    if stable != True:
        time = stable
        return time

    tseries = np.array(tseries)
    alltime = []

    Xs = []
    for i, trio in enumerate(trios):
        # These are the .npy.
        cur_tseries = tseries[None, i, :].astype(np.float32)
        mass_array = np.array([sim.particles[j].m/sim.particles[0].m for j in trio]).astype(np.float32)
        mass_array = E.repeat(mass_array, 'i -> () t i', t=100)
        X = data_setup_kernel(mass_array, cur_tseries)
        Xs.append(X)

    Xs = np.array(Xs)
    return Xs

def fast_truncnorm(
        loc, scale, left=np.inf, right=np.inf,
        d=10000, nsamp=50):
    """Fast truncnorm sampling.
    
    Assumes scale and loc have the desired shape of output.
    length is number of elements.
    Select nsamp based on expecting at minimum one sample of a Gaussian
        to fit within your (left, right) range.
    Select d based on memory considerations - need to operate on
        a (d, nsamp) array.
    """
    oldscale = scale
    oldloc = loc
    
    scale = scale.reshape(-1)
    loc = loc.reshape(-1)
    t_inst_samples = np.zeros_like(scale)
    start = 0
        
    for start in range(0, scale.shape[0], d):

        end = start + d
        if end > scale.shape[0]:
            end = scale.shape[0]
        
        cd = end-start
        rand_out = np.random.normal(size=(nsamp, cd))

        rand_out = (
            rand_out * scale[None, start:end]
            + loc[None, start:end]
        )
        
        #rand_out is (nsamp, cd)
        if right == np.inf:
            mask = (rand_out > left)
        elif left == np.inf:
            mask = (rand_out < right)
        else:
            mask = (rand_out > left) & (rand_out < right)
            
        first_good_val = rand_out[
            mask.argmax(0), np.arange(cd)
        ]
        t_inst_samples[start:end] = first_good_val
        
    return t_inst_samples.reshape(*oldscale.shape)

class DeepRegressor(object):
    def __init__(self, cuda=False, filebase='*v50_*output.pkl'):
        super(DeepRegressor, self).__init__()
        pwd = os.path.dirname(__file__)
        pwd = pwd + '/models/regression'
        self.cuda = cuda

        #Load model
        self.swag_ensemble = [
            load_swag(fname).cpu()
            for i, fname in enumerate(glob.glob(pwd + '/' + filebase)) #0.78, 0.970
        ]
        #Data scaling parameters:
        self.scale_ = np.array([2.88976974e+03, 6.10019661e-02, 4.03849732e-02, 4.81638693e+01,
                   6.72583662e-02, 4.17939679e-02, 8.15995339e+00, 2.26871589e+01,
                   4.73612029e-03, 7.09223721e-02, 3.06455099e-02, 7.10726478e-01,
                   7.03392022e-01, 7.07873597e-01, 7.06030923e-01, 7.04728204e-01,
                   7.09420909e-01, 1.90740659e-01, 4.75502285e-02, 2.77188320e-02,
                   7.08891412e-01, 7.05214134e-01, 7.09786887e-01, 7.04371833e-01,
                   7.04371110e-01, 7.09828420e-01, 3.33589977e-01, 5.20857790e-02,
                   2.84763136e-02, 7.02210626e-01, 7.11815232e-01, 7.10512240e-01,
                   7.03646004e-01, 7.08017286e-01, 7.06162814e-01, 2.12569430e-05,
                   2.35019125e-05, 2.04211110e-05, 7.51048890e-02, 3.94254400e-01,
                   7.11351099e-02])
        self.mean_ = np.array([ 4.95458585e+03,  5.67411891e-02,  3.83176945e-02,  2.97223474e+00,
                   6.29733979e-02,  3.50074471e-02,  6.72845676e-01,  9.92794768e+00,
                   9.99628430e-01,  5.39591547e-02,  2.92795061e-02,  2.12480714e-03,
                  -1.01500319e-02,  1.82667162e-02,  1.00813201e-02,  5.74404197e-03,
                   6.86570242e-03,  1.25316320e+00,  4.76946516e-02,  2.71326280e-02,
                   7.02054326e-03,  9.83378673e-03, -5.70616748e-03,  5.50782881e-03,
                  -8.44213953e-04,  2.05958338e-03,  1.57866569e+00,  4.31476211e-02,
                   2.73316392e-02,  1.05505555e-02,  1.03922250e-02,  7.36865006e-03,
                  -6.00523246e-04,  6.53016990e-03, -1.72038113e-03,  1.24807860e-05,
                   1.60314173e-05,  1.21732696e-05,  5.67292645e-03,  1.92488263e-01,
                   5.08607199e-03])

    def sample_full_swag(self, X_sample):
        """Pick a random model from the ensemble and sample from it
            within each model, it samples from its weights."""
        
        swag_i = np.random.randint(0, len(self.swag_ensemble))
        swag_model = self.swag_ensemble[swag_i]
        swag_model.eval()
        if self.cuda:
            swag_model.w_avg = swag_model.w_avg.cuda()
            swag_model.w2_avg = swag_model.w2_avg.cuda()
            swag_model.pre_D = swag_model.pre_D.cuda()
            swag_model.cuda()
        out = swag_model.forward_swag_fast(X_sample, scale=0.5)
        if self.cuda:
            swag_model.w_avg = swag_model.w_avg.cpu()
            swag_model.w2_avg = swag_model.w2_avg.cpu()
            swag_model.pre_D = swag_model.pre_D.cpu()
            swag_model.cpu()
        return out

    def predict_instability_time(self, sim, samples=1000, seed=None,
            max_model_samples=100, return_samples=False, prior_above_9=fitted_prior(), Ncpus=None):
        """Estimate instability time for given simulation(s), and the 68% confidence
            interval.

        Returns the median of the posterior, the 16th percentile, and
            the 84th percentile. Uses `samples` samples of the posterior
            to calculate this.

        Parameters:

        sim (rebound.Simulation or list): Orbital configuration(s) to test
        samples (int): Number of samples to use
        seed (int): Random seed
        max_model_samples (int): maximum number of times to re-generate model parameters.
            Larger number increases accuracy but greatly decreases speed.
        return_samples (bool): return the raw samples as a second argument
        prior_above_9 (function): function defining the probability density
            function of instability times above 1e9 orbits of the innermost
            planet. By default is a decaying prior which was fit to the training dataset.
            This takes as input time in terms of the orbits of the innermost planet.
        Ncpus (int): Number of CPUs to use for calculation (only if passing more than one simulation).
            Default: Use all available cpus. 

        Returns:

        center_estimate (float): instability time in units of
            the rebound simulation's time units (e.g., if P=1.
            for the innermost planet, this estimate will be
            in units of orbits)
        lower (float): 16th percentile instability time
        upper (float): 84th percentile instability time
        [t_inst_samples (array): raw samples of the posterior]
        """
        batched = self.is_batched(sim)
        t_inst_samples = self.sample_instability_time(sim,
                samples=samples, seed=seed, max_model_samples=max_model_samples,
                prior_above_9=prior_above_9, Ncpus=Ncpus)
        if batched:
            center_estimate = np.median(t_inst_samples, axis=1)
            upper = np.percentile(t_inst_samples, 100-16, axis=1)
            lower = np.percentile(t_inst_samples,     16, axis=1)
        else:
            center_estimate = np.median(t_inst_samples)
            upper = np.percentile(t_inst_samples, 100-16)
            lower = np.percentile(t_inst_samples,     16)

        if return_samples:
            return center_estimate, lower, upper, t_inst_samples
        else:
            return center_estimate, lower, upper

    def predict_stable(self, sim, tmax=None, samples=1000, seed=None, 
            return_samples=False, max_model_samples=100, prior_above_9=fitted_prior(), Ncpus=None):
        """Estimate chance of stability for given simulation(s).

        Parameters:

        sim (rebound.Simulation or list): Orbital configuration(s) to test
        tmax (float or list): Time at which the system is queried as stable,
            in rebound simulation time units.
        samples (int): Number of samples to use
        seed (int): Random seed
        max_model_samples (int): maximum number of times to re-generate model parameters.
            Larger number increases accuracy but greatly decreases speed.
        return_samples (bool): return the raw samples as a second argument
        prior_above_9 (function): function defining the probability density
            function of instability times above 1e9 orbits of the innermost
            planet. By default is a decaying prior which was fit to the training dataset.
            This takes as input time in terms of the orbits of the innermost planet.
        Ncpus (int): Number of CPUs to use for calculation (only if passing more than one simulation).
            Default: Use all available cpus. 

        Returns:

        p (float): probability of stability past the given tmax
            (default 1e9*min(P) orbits)
        [t_inst_samples (array): raw samples of the posterior]
        """
        batched = self.is_batched(sim)
        t_inst_samples = self.sample_instability_time(sim,
                samples=samples, seed=seed, max_model_samples=max_model_samples,
                prior_above_9=prior_above_9)

        if tmax is None:
            if batched:
                tmax = np.array([
                    1e9 *  
                    np.min([np.abs(p.P) for p in s.particles[1:s.N_real]])
                    for s in sim])
            else:
                minP = np.min([np.abs(p.P) for p in sim.particles[1:sim.N_real]])
                tmax = 1e9 * minP
        elif batched:
            if isinstance(tmax, list):
                tmax = np.array(tmax)
            elif isinstance(tmax, np.ndarray):
                ...
            else:
                tmax = np.ones(len(sim)) * tmax
            assert len(tmax) == len(sim)

        if batched:
            out = np.average(t_inst_samples[:, :] > tmax[:, None], 1)
        else:
            out = np.average(t_inst_samples > tmax)
        
        if return_samples:
            return out, t_inst_samples
        else:
            return out

    def resample_stable_sims(self, samps_time, prior_above_9):
        """Use a prior to re-sample stable instability times"""
        stable_past_9 = samps_time >= 9
        normalization = quad(prior_above_9, a=9, b=np.inf)[0]
        prior = lambda logT: prior_above_9(logT)/normalization
        n_samples = stable_past_9.sum()
        bins = max([10000, n_samples*4])
        top = 100.
        bin_edges = np.linspace(9, top, num=bins)
        cum_values = [0] + list(np.cumsum(prior(bin_edges)*(bin_edges[1] - bin_edges[0]))) + [1]
        bin_edges = [9.] +list(bin_edges)+[top]
        # Numerically interpolate the inverse cumulative distribution function:
        inv_cdf = interp1d(cum_values, bin_edges)
        r = np.random.rand(n_samples)
        t_inst_samples = inv_cdf(r)
        samps_time[stable_past_9] = t_inst_samples
        return samps_time

    def is_batched(self, sim):
        batched = False
        if isinstance(sim, list):
            batched = True
            nsim = len(sim)
            if len(set([s.N_real for s in sim])) != 1:
                raise ValueError("If running over many sims at once, they must have the same number of particles!")
        return batched

    @profile
    def sample_instability_time(self, sim, samples=1000, seed=None,
            max_model_samples=100, prior_above_9=fitted_prior(), Ncpus=None):
        """Return samples from a posterior over instability time for
            given simulation(s). This returns samples from a simple prior for
            all times greater than 10^9 orbits.

        Parameters:

        sim (rebound.Simulation or list): Orbital configuration(s) to test
        samples (int): Number of samples to return
        seed (int): Random seed
        max_model_samples (int): maximum number of times to re-generate model parameters.
            Larger number increases accuracy but greatly decreases speed.
        prior_above_9 (function): function defining the probability density
            function of instability times above 1e9 orbits of the innermost
            planet. By default is a decaying prior which was fit to the training dataset.
            This takes as input time in terms of the orbits of the innermost planet.
        Ncpus (int): Number of CPUs to use for calculation (only if
            passing more than one simulation). Default: Use all available cpus. 

        Returns:

        np.array: samples of the posterior (nsamples,) or (nsim, nsamples) for
            instability time, in units of the rebound simulation.
        """
        batched = self.is_batched(sim)

        if seed is not None:
            os.environ["PL_GLOBAL_SEED"] = str(seed)
            random.seed(seed)
            np.random.seed(seed)
            torch.manual_seed(seed)
            if self.cuda:
                torch.cuda.manual_seed_all(seed)

        if batched:
            n_sims = len(sim)
            if Ncpus is None:
                Ncpus = cpu_count()
            with Pool(Ncpus) as pool:
                pool_out = pool.map(generate_dataset, sim)

            Xs = np.array([X for X in pool_out if isinstance(X, np.ndarray)])
            already_computed_results_idx =   [i for i, X in enumerate(pool_out) if not isinstance(X, np.ndarray)]
            already_computed_results_times = [X for i, X in enumerate(pool_out) if not isinstance(X, np.ndarray)]
        else:
            out = generate_dataset(sim)
            if not isinstance(out, np.ndarray):
                minP = np.min([np.abs(p.P) for p in sim.particles[1:sim.N_real]])
                return np.ones(samples) * out * minP
            Xs = np.array([out])

        if len(Xs) > 0:
            nbatch = Xs.shape[0]
            ntrios = Xs.shape[1]
            nt = 100
            X = E.rearrange(Xs, 'batch trio () time feature -> (batch trio time) feature')
            Xp = (X - self.mean_[None, :]) / self.scale_[None, :]
            Xp = E.rearrange(Xp, '(batch trio time) feature -> (batch trio) time feature',
                             batch=nbatch, trio=ntrios, time=nt)

            Xflat = torch.tensor(Xp).float()
            if self.cuda:
                Xflat = Xflat.cuda()

            model_samples = min([max_model_samples, samples])
            oversample = int(math.ceil(samples/model_samples))
            Xflat = E.repeat(Xflat,
                            '(batch trio) time feature -> (batch trio oversample) time feature',
                            batch=nbatch, trio=ntrios,
                            oversample=oversample)

            sampled_mu_std = np.array([self.sample_full_swag(Xflat).detach().cpu().numpy() for _ in range(model_samples)])
            sampled_mu_std = E.rearrange(sampled_mu_std,
                            'msamples (batch trio oversample) mu_std -> (msamples oversample) (batch trio) mu_std',
                            batch=nbatch, trio=ntrios,
                            oversample=oversample
                            )
            sampled_mu_std = sampled_mu_std[:samples]
            sampled_mu_std = sampled_mu_std.astype(np.float64)
            # print(sampled_mu_std.shape)
            #(samples, (batch trio), mu_std)

            samps_time = np.array(fast_truncnorm(
                    sampled_mu_std[..., 0], sampled_mu_std[..., 1],
                    left=4, d=10000, nsamp=40
                ))
            samps_time = self.resample_stable_sims(samps_time, prior_above_9)
            samps_time = E.rearrange(samps_time,
                             'samples (batch trio) -> batch samples trio',
                             batch=nbatch, trio=ntrios)
            outs = np.min(samps_time, 2)
            time_estimates = np.power(10.0, outs)
            # print(time_estimates.shape)
            #HACK TODO - need already computed estimates

        if not batched:
            minP = np.min([np.abs(p.P) for p in sim.particles[1:sim.N_real]])
            return time_estimates[0] * minP

        j = 0
        k = 0
        correct_order_results = []
        for i in range(n_sims):
            cur_sim = sim[i]
            minP = np.min([np.abs(p.P) for p in cur_sim.particles[1:cur_sim.N_real]])
            if i in already_computed_results_idx:
                correct_order_results.append(already_computed_results_times[k] * np.ones(samples) * minP)
                k += 1
            else:
                correct_order_results.append(time_estimates[j] * minP)
                j += 1

        correct_order_results = np.array(correct_order_results, dtype=np.float64)
        return correct_order_results

@profile
def data_setup_kernel(mass_array, cur_tseries):
    """Data preprocessing"""
    nf = 32
    nt = 100
    concat_with_mass = np.concatenate((cur_tseries, mass_array), axis=2)
    assert concat_with_mass.shape == (1, nt, nf - 3)

    concat_with_nan = np.concatenate(
            (concat_with_mass,
            (~np.isfinite(concat_with_mass[..., [3, 6, 7]]))),
        axis=2)

    clean_input = np.nan_to_num(concat_with_nan, posinf=0.0, neginf=0.0)

    X = np.zeros((1, nt, nf + 9))

    cur_feature = 0
    for j in range(nf):
        if j in [11, 12, 13, 17, 18, 19, 23, 24, 25]: #if 'Omega' in label or 'pomega' in label or 'theta' in label:
            X[:, :, cur_feature] = np.cos(clean_input[:, :, j])
            cur_feature += 1
            X[:, :, cur_feature] = np.sin(clean_input[:, :, j])
            cur_feature += 1
        else:
            X[:, :, cur_feature] = clean_input[:, :, j]
            cur_feature += 1

    return X



# -*- coding: utf-8 -*-
"""Stability of Planetary Orbital Configurations Klassifier"""

from .featureclassifier import FeatureClassifier
from .deepregressor import DeepRegressor
from .nbodyregressor import NbodyRegressor
from .analyticalclassifier import AnalyticalClassifier
from .version import __version__

__all__ = ["FeatureClassifier", "DeepRegressor", "NbodyRegressor", "AnalyticalClassifier"]



import rebound
import pandas as pd
if rebound.__githash__ != '6fb912f615ca542b670ab591375191d1ed914672':
    print('Check out rebound commit 6fb912f615ca542b670ab591375191d1ed914672 and rerun script. See spock/training.md')
    [
datapath = '../data/'

randompath = 'random/simulation_archives/runs/'
respath = 'resonant/simulation_archives/runs/'

dfrand = pd.read_csv('../training_data/random/labels.csv', index_col=0)
dfres = pd.read_csv('../training_data/resonant/labels.csv', index_col=0)

def randPratios(row):
    sa = rebound.SimulationArchive(datapath+randompath+'sa{0}'.format(row['runstring']))
    sim = sa[0]
    ps = sim.particles
    row['Pratio21'] = ps[2].P/ps[1].P
    row['Pratio32'] = ps[3].P/ps[2].P
    RH12 = ps[1].a*((ps[1].m+ps[2].m)/3.)**(1/3)
    RH23 = ps[2].a*((ps[2].m+ps[3].m)/3.)**(1/3)
    row['beta12'] = (ps[2].a-ps[1].a)/RH12
    row['beta23'] = (ps[3].a-ps[2].a)/RH23
    return row

def resPratios(row):
    sa = rebound.SimulationArchive(datapath+respath+'sa{0}'.format(row['runstring']))
    sim = sa[0]
    ps = sim.particles
    row['Pratio21'] = ps[2].P/ps[1].P
    row['Pratio32'] = ps[3].P/ps[2].P
    RH12 = ps[1].a*((ps[1].m+ps[2].m)/3.)**(1/3)
    RH23 = ps[2].a*((ps[2].m+ps[3].m)/3.)**(1/3)
    row['beta12'] = (ps[2].a-ps[1].a)/RH12
    row['beta23'] = (ps[3].a-ps[2].a)/RH23
    return row

dfrand = dfrand.apply(randPratios, axis=1)
dfrand.to_csv('randomPratios.csv')

dfres = dfres.apply(resPratios, axis=1)
dfres.to_csv('resPratios.csv')



import rebound
import numpy as np
from celmech.secular import LaplaceLagrangeSystem
from celmech import Poincare

class AnalyticalClassifier():
    def __init__(self):
        pass
    def check_errors(self, sim):
        if sim.N_real < 4:
            raise AttributeError("SPOCK Error: SPOCK only applicable to systems with 3 or more planets") 
        
    def _eminus_max(self, lsys, Lambda, i1, i2):
        if i1 > i2:
            raise AttributeEror("i2 must be exterior body")
        res = {}
        res['L10'] = Lambda[i1-1]
        res['L20'] = Lambda[i2-1]
        res['L12'] = res['L10'] + res['L20']
        res['Lr'] = res['L10']*res['L20']/res['L12']
        res['psi'] = np.pi + np.arcsin(np.sqrt(res['L10']/res['L12']))
        res['R'] = np.identity(Lambda.shape[0]) # Nplanets x Nplanets rotation matrix
        # Only rotate the 2x2 submatrix corresponding to i1 and i2 with rotation matrix
        res['R'][np.ix_([i1-1,i2-1],[i1-1, i2-1])] = np.array([[np.cos(res['psi']), -np.sin(res['psi'])], [np.sin(res['psi']), np.cos(res['psi'])]])

        Mnorm = res['R'] @ lsys.Neccentricity_matrix @ res['R'].T
        Fx0 = res['R'] @ lsys.kappa0_vec/res['Lr']**(1/2)
        Fy0 = res['R'] @ lsys.eta0_vec/res['Lr']**(1/2)

        vals,T = np.linalg.eigh(Mnorm)

        Ax0 = T.T @ Fx0
        Ay0 = T.T @ Fy0
        A = np.sqrt(Ax0**2 + Ay0**2) # Mode amplitudes (A[0], A[1])
        Fmax = np.abs(T) @ A.T
        return Fmax[i1-1] # i1-1 index is eminus for the i1,i2 pair

    def _calc_tau_pair(self, sim, lsys, Lambda, i1, i2):
        if i1 > i2:
            raise AttributeEror("i2 must be exterior body")
        ps = sim.particles
        delta = (ps[i2].a-ps[i1].a)/ps[i2].a/(ps[i1].m+ps[i2].m)**(1/4)
        emax = self._eminus_max(lsys, Lambda, i1, i2)
        ec = (1-(ps[i1].P/ps[i2].P)**(2/3))
        tau = (1.8/delta)**2/np.abs(np.log(emax/ec))**(3/2)
        return tau

    def _calc_tau(self, sim):
        lsys = LaplaceLagrangeSystem.from_Simulation(sim)
        pvars = Poincare.from_Simulation(sim)
        Lambda = np.array([p.Lambda for p in pvars.particles[1:]])

        tau_max = 0
        for i in range(1, sim.N_real):
            tau = 0
            if i-1 >= 1:
                tau += self._calc_tau_pair(sim, lsys, Lambda, i-1, i)
            if i+1 < sim.N_real:
                tau += self._calc_tau_pair(sim, lsys, Lambda, i, i+1)
            if tau > tau_max:
                tau_max = tau
        return tau_max

    def calc_tau(self, sim):
        if isinstance(sim, rebound.Simulation):
            sim = [sim]
        Nsims = len(sim)
        taus = [self._calc_tau(s) for s in sim]

        if Nsims == 1:
            return taus[0]
        else:
            return taus

    def predict_stable(self, sim, n_jobs=-1):
        """
        Predict whether passed simulation will be stable over 10^9 orbits of the innermost planet.

        Parameters:

        sim (rebound.Simulation): Orbital configuration to test
        n_jobs (int):               Number of cores to use for calculation (only if passing more than one simulation). Default: Use all available cores. 

        Returns:

        float:  Estimated probability of stability. Will return exactly zero if configuration goes 
                unstable within first 10^4 orbits.

        """
        if isinstance(sim, rebound.Simulation):
            sim = [sim]
        Nsims = len(sim)
        
        tau = self.calc_tau(sim)
        prob = np.maximum(1-tau, 0) 
        
        return prob



import numpy as np
import rebound
from .simsetup import init_sim_parameters
from multiprocessing.pool import ThreadPool
from multiprocessing import cpu_count

class NbodyRegressor():
    def __init__(self):
        pass 
    
    def predict_instability_time(self, sim, tmax=None, archive_filename=None, archive_interval=None, n_jobs=-1, match_training=False):
        """
        Predict instability time through N-body integration.

        Parameters:

        sim (rebound.Simulation, or list):   Orbital configuration(s) to test
        tmax (float):               Maximum time to integrate for (in Simulation time units). If passing a list of sims, need to pass a list of tmax of equal length. Defaults to 1e9 innermost planet orbits.
        archive_filename (str):     Path and filename to store a rebound SimulationArchive of snapshots
        archive_interval (float):   Time between snapshots for SimulationArchive (in Simulation time units)
        n_jobs (int):               Number of cores to use for calculation (only if passing more than one simulation). Default: Use all available cores. 
        match_training (bool):      In order to match the exact chaotic trajectory given an input orbital configuration from our training set, need to set match_training=True. False gives an equally valid chaotic realization, and gives a factor of ~2 speedup in evaluation.

        Returns:

        float:  Time of instability (Hill sphere crossing), in Simulation time units, or tmax if stable
        int:    1 if integration reached tmax, 0 if Hill spheres crossed
        """
 
        if isinstance(sim, rebound.Simulation):
            sim = [sim]
            tmax = [tmax]
        else:
            if tmax is None:
                tmax = [None]*len(sim)
            if len(tmax) != len(sim):
                raise ValueError("When passing multiple simulations, need to pass a list for tmax of the same length (or not pass tmax to default to 10^9 innermost planet orbits for all simulations)")

        args = []
        for i, s in enumerate(sim):
            s = s.copy()
            if match_training == True:
                init_sim_parameters(s)
            else:
                init_sim_parameters(s, megno=False, safe_mode=0)
            minP = np.min([p.P for p in s.particles[1:s.N_real]])
            if tmax[i] is None:
                tmax[i] = 1e9*minP
            if archive_filename:
                if archive_interval is None:
                    archive_interval = tmax[i]/1000
                if len(sim) == 1: # single sim
                    s.automateSimulationArchive(archive_filename, archive_interval, deletefile=True)
                else: # strip final extension and add an index for each simarchive
                    ext = archive_filename.split('.')[-1]
                    len_ext = len(ext)+1
                    filename = "{0}_{1}.{2}".format(archive_filename[:-len_ext], i, ext)
                    s.automateSimulationArchive(filename, archive_interval, deletefile=True)
            args.append([s, tmax[i]])
        
        def run(params):
            sim, tmax = params
            if np.isnan(sim.dt): # sim.dt set to nan for hyperbolic initial conditions in simsetup/set_integrator_and_timestep
                return np.nan
            try:
                sim.integrate(tmax, exact_finish_time=0)
            except (rebound.Collision, rebound.Escape):
                if sim._simulationarchive_filename:
                    sim.simulationarchive_snapshot(sim._simulationarchive_filename.decode('utf-8'))
                return sim.t
           
            return tmax

        if len(args) == 1: # single sim
            tinst = run(args[0])
        else:
            if n_jobs == -1:
                n_jobs = cpu_count()
            pool = ThreadPool(n_jobs)
            tinst = np.array(pool.map(run, args))
        
        # Uncertainty estimates come from Hussain & Tamayo 2020. Extra factor of sqrt(2) comes from fact that in addition to having a spread in instability times from chaos, we don't know where the mean is from a single integration. See paper
        lower = 10**(np.log10(tinst)-np.sqrt(2)*0.43)
        upper = 10**(np.log10(tinst)+np.sqrt(2)*0.43)

        if len(args) == 1:
            if tinst == tmax[0]:
                lower = 0
                upper = 0
        else: # set confidence intervals for all sims that reached tmax to zero 
            tmax = np.array(tmax)
            lower[tinst == tmax] = 0
            upper[tinst == tmax] = 0 

        return tinst, lower, upper

    def predict_stable(self, sim, tmax=None, archive_filename=None, archive_interval=None, n_jobs=-1, match_training=False):
        """
        Predict whether system is stable up to time=tmax through N-body integration.

        Parameters:

        sim (rebound.Simulation):   Orbital configuration to test
        tmax (float):               Maximum time to integrate for (in Simulation time units) 
        archive_filename (str):     Path and filename to store a rebound SimulationArchive of snapshots
        archive_interval (float):   Time between snapshots for SimulationArchive (in Simulation time units)
        n_jobs (int):               Number of cores to use for calculation (only if passing more than one simulation). Default: Use all available cores. 
        match_training (bool):      In order to match the exact chaotic trajectory given an input orbital configuration from our training set, need to set match_training=True. False gives an equally valid chaotic realization, and gives a factor of ~2 speedup in evaluation.

        Returns:

        float:  Time of instability (Hill sphere crossing), in Simulation time units, or tmax if stable
        int:    1 if integration reached tmax, 0 if Hill spheres crossed
        """
 
        tinst, lower, upper = self.predict_instability_time(sim, tmax, archive_filename, archive_interval, n_jobs, match_training) 
        # If tmax == None, each sim can have different tmaxs
        # Use fact that if confidence intervals == 0, we hit tmax limit
        # (lower = np.nan if error, >0 if hit instability)
        stable = (lower == 0) 
        try:
            return stable.astype(int)
        except:
            return int(stable) # single simulation



import numpy as np
import unittest
import rebound
from spock import NbodyRegressor
from spock.simsetup import init_sim_parameters
from spock.feature_functions import features

def unstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-4, P=1)
    sim.add(m=1.e-4, P=1.3)
    sim.add(m=1.e-4, P=1.6)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim

def longstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-7, P=1)
    sim.add(m=1.e-7, P=2.1)
    sim.add(m=1.e-7, P=4.5)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim

def solarsystemsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.7e-7, a=0.39, e=0.21)
    sim.add(m=2.4e-6, a=0.72, e=0.007)
    sim.add(m=3.e-6, a=1, e=0.017)
    sim.add(m=3.2e-7, a=1.52, e=0.09)
    sim.add(m=1.e-3, a=5.2, e=0.049)
    sim.add(m=2.9e-4, a=9.54, e=0.055)
    sim.add(m=4.4e-5, a=19.2, e=0.047)
    sim.add(m=5.2e-5, a=30.1, e=0.009)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim
        
def hyperbolicsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-5, a=-1., e=1.2)
    sim.add(m=1.e-5, a=2.)
    sim.add(m=1.e-5, a=3.)
    return sim

def escapesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-12, P=3.14, e=0.03, l=0.5)
    sim.add(m=1.e-12, P=4.396, e=0.03, l=4.8)
    sim.add(m=1.e-12, a=100, e=0.999)
    return sim

def vstablesim():
    sim = rebound.Simulation()
    sim.add(m=1)
    sim.add(m=1e-7, P=1.)
    sim.add(m=1e-7, P=1.8)
    sim.add(m=1e-7, P=3.2)
    return sim

def rescale(sim, dscale, tscale, mscale):                                                                      
    simr = rebound.Simulation()
    vscale = dscale/tscale 
    simr.G *= mscale*tscale**2/dscale**3

    for p in sim.particles:
        simr.add(m=p.m/mscale, x=p.x/dscale, y=p.y/dscale, z=p.z/dscale, vx=p.vx/vscale, vy=p.vy/vscale, vz=p.vz/vscale, r=p.r/dscale)

    return simr

class TestNbody(unittest.TestCase):
    def setUp(self):
        self.model = NbodyRegressor()

    def test_list_stable(self):
        tmax = [1e4, 1e4, 1, 1e4, 1e4] # test that unstablesim in middle still classified as stable with tmax=1
        stable_target = [0, 0, 1, 1, 1]
        stable = self.model.predict_stable([hyperbolicsim(), escapesim(), unstablesim(), longstablesim(), solarsystemsim()], tmax=tmax)
        self.assertSequenceEqual(stable.tolist(), stable_target)
    
    def test_mismatched_lists(self):
        tmax = [1e4, 1e4, 1] # test that unstablesim in middle still classified as stable with tmax=1
        with self.assertRaises(ValueError):
            stable = self.model.predict_stable([hyperbolicsim(), escapesim(), unstablesim(), longstablesim()], tmax=tmax)

    def test_list_time(self):
        tmax = [1e4, 1e4, 1, 1e4, 1e4] # test that unstablesim in middle still classified as stable with tmax=1
        tinst, lower, upper = self.model.predict_instability_time([hyperbolicsim(), escapesim(), unstablesim(), longstablesim(), solarsystemsim()], tmax=tmax)
        self.assertTrue(np.isnan(tinst[0]))
        self.assertLess(tinst[1], tmax[1])
        for i in range(2,5):
            self.assertEqual(tinst[i], tmax[i])

    def test_repeat(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        p1 = self.model.predict_stable(sim, tmax=1e4)
        p2 = self.model.predict_stable(sim, tmax=1e4)
        self.assertEqual(p1, p2)

    def test_hyperbolic(self):
        sim = hyperbolicsim()
        self.assertEqual(self.model.predict_stable(sim, tmax=1e4), 0)
    
    def test_escape(self):
        sim = escapesim()
        self.assertEqual(self.model.predict_stable(sim, tmax=1e4), 0)
    
    def test_unstable_in_short_integration(self):
        sim = unstablesim()
        self.assertEqual(self.model.predict_stable(sim, tmax=1e4), 0)
    
    def test_solarsystem(self):
        sim = solarsystemsim()
        self.assertEqual(self.model.predict_stable(sim, tmax=1e4), 1)
    
    def test_stable(self):
        sim = longstablesim()
        self.assertEqual(self.model.predict_stable(sim, tmax=1e4), 1)

    def test_same_traj_as_feature_classifier(self):
        # procedure in featureclassifier
        sim = unstablesim()
        init_sim_parameters(sim)                            
        trios = [[i,i+1,i+2] for i in range(1,sim.N_real-2)]
        featureargs = [10000, 80, trios]                    
        triofeatures, stable = features(sim, featureargs)   
        tfc = sim.t

        sim = unstablesim()
        init_sim_parameters(sim)
        try:
            sim.integrate(1e4*sim.particles[1].P, exact_finish_time=0)
        except rebound.Collision:
            tn = sim.t
        self.assertEqual(tfc, tn)

if __name__ == '__main__':
    unittest.main()



import rebound
import numpy as np
import pandas as pd
import os
from subprocess import call
import warnings
warnings.filterwarnings('ignore') # filter REBOUND warnings about version that I've already tested

datapath = '../data/'
repopath = '../'

if rebound.__githash__ != '6fb912f615ca542b670ab591375191d1ed914672':
    print('Need to install REBOUND commit 6fb912f615ca542b670ab591375191d1ed914672 to ensure this runs correctly')
    exit()

def labels(row):
    try:
        sa = rebound.SimulationArchive(pathtosa+'sa'+row['runstring'])
        sim = sa[0]
        P1 = sim.particles[1].P # Need initial orbital period for TTVsystems, where P1 != 1

        try: # Needed for old integrations (random and Naireen) because no snapshot on end
            sim = rebound.Simulation(pathtosa+'../../final_conditions/runs/fc'+row['runstring'])
            row['instability_time'] = sim.t/P1
            try:
                ssim = rebound.Simulation(pathtossa+'../../final_conditions/shadowruns/fc'+row['runstring'])
                row['shadow_instability_time'] = ssim.t/P1
            except:
                print('No shadow for random {0}'.format(row['runstring']))
        except: # New runs (resonant and Ari) have snapshots at collision
            sim = sa[-1]
            if sim.t > 9.99e3 and sim.t < 1.0001e4: # catch all stable integrations accidentally thrown out. Have checked this catches all thrown out, and doesn't catch any with actual  instability times in this range
                row['instability_time'] = 1e9
                row['shadow_instability_time'] = 1e9
            else:
                row['instability_time'] = sim.t/P1
                ssa = rebound.SimulationArchive(pathtossa+'sa'+row['runstring'])
                ssim = ssa[-1]
                row['shadow_instability_time'] = ssim.t/P1
        row['Stable'] = row['instability_time'] > 9.99e8
    except:
        print('Error', pathtosa+'sa'+row['runstring'])
    return row

def massratios(row):
    try:
        sa = rebound.SimulationArchive(pathtosa+'sa'+row['runstring'])
        sim = sa[0]
        row['m1'] = sim.particles[1].m/sim.particles[0].m
        row['m2'] = sim.particles[2].m/sim.particles[0].m
        row['m3'] = sim.particles[3].m/sim.particles[0].m
    except:
        print(pathtosa+'sa'+row['runstring'])
    return row

def ttvsystems():
    folders = ['KOI-1576']
    return ['TTVsystems/' + folder for folder in folders]

def nonressystems():
    folders = ['Kepler-431']
    return ['nonressystems/' + folder for folder in folders]


datasets = ['resonant', 'random'] + ttvsystems() + nonressystems()
for dataset in datasets:
    print(dataset)
    pathtosa = datapath + dataset + '/simulation_archives/runs/'
    pathtossa = datapath + dataset + '/simulation_archives/shadowruns/'
    pathtotraining = repopath + 'training_data/' + dataset + '/'

    root, dirs, files = next(os.walk(pathtosa))
    runstrings = [file[2:] for file in files if file[-3:] == 'bin']
    df = pd.DataFrame(runstrings, columns=['runstring'])
    df = df.sort_values(by='runstring')
    df = df.reset_index(drop=True)
    df.to_csv(pathtotraining+'runstrings.csv', encoding='ascii')

    df['instability_time'] = -1
    df['shadow_instability_time'] = -1
    df['Stable'] = -1

    df = df.apply(labels, axis=1)
    df.to_csv(pathtotraining+'labels.csv', encoding='ascii')
    df = pd.DataFrame(df['shadow_instability_time'])
    call('mkdir ' + pathtotraining + 'shadowtimes', shell=True)
    call('cp ' + pathtotraining + 'labels.csv ' + pathtotraining + 'shadowtimes/', shell=True)
    df.to_csv(pathtotraining+'shadowtimes/trainingdata.csv', encoding='ascii')

    df = pd.read_csv(pathtotraining+'runstrings.csv', index_col=0)
    df['m1'] = -1
    df['m2'] = -1
    df['m3'] = -1

    df = df.apply(massratios, axis=1)
    df.to_csv(pathtotraining+'massratios.csv', encoding='ascii')
    call('cp ' + pathtotraining + 'massratios.csv ' + pathtotraining + 'shadowtimes/', shell=True)
    call('cp ' + pathtotraining + 'runstrings.csv ' + pathtotraining + 'shadowtimes/', shell=True)



import rebound
import numpy as np
from scipy.optimize import brenth

def F(e,alpha,gamma):
    """Equation 35 of Laskar & Petit (2017)"""
    denom = np.sqrt(alpha*(1-e*e)+gamma*gamma*e*e)
    return alpha*e -1 + alpha + gamma*e / denom

### start AMD functions

def AMD_crit(sim2, i1, i2): # assumes i1.a < i2.a
    try:
        sim = sim2.copy()
        sim.move_to_com() # AMD calculations easiest in canonical heliocentric coordinates, but don't want to change original sim so copy
    except: # old version of REBOUND doesn't have a copy function (for random dataset), but all those were already moved to com so OK
        sim = sim2
    # returns AMD_crit dimensionalized to Lambdaout
    ps = sim.particles
    if ps[i1].m == 0. or ps[i2].m == 0:
        return 0 # if one particle is massless, any amount of AMD can take it to e=1, so AMDcrit = 0 (always AMD unstable)

    mu = sim.G*ps[0].m
    alpha = ps[i1].a / ps[i2].a
    gamma = ps[i1].m / ps[i2].m
    LambdaPrime = ps[i2].m * np.sqrt(mu*ps[i2].a)
    curlyC = relative_AMD_crit(alpha,gamma)
    AMD_crit = curlyC * LambdaPrime # Eq 29 AMD_crit = C = curlyC*Lambda'

    return AMD_crit

def relative_AMD_crit(alpha,gamma):
    """Equation 29"""
    e0 = np.min((1,1/alpha-1))
    try:
        ec = brenth(F,0,e0,args=(alpha,gamma))
    except: # can fail to converge
        print('brenth failed: e0  = {0}, alpha = {1}, gamma= {2}'.format(e0, alpha, gamma))
        return np.nan
    e1c = np.sin(np.arctan(gamma*ec / np.sqrt(alpha*(1-ec*ec))))
    curlyC = gamma*np.sqrt(alpha) * (1-np.sqrt(1-ec*ec)) + (1 - np.sqrt(1-e1c*e1c))
    return curlyC

def AMD(sim2):
    try:
        sim = sim2.copy()
        sim.move_to_com() # AMD calculations easiest in canonical heliocentric coordinates, but don't want to change original sim so copy
    except: # old version of REBOUND doesn't have a copy function (for random dataset), but all those were already moved to com so OK
        sim = sim2
    ps = sim.particles
    Lx, Ly, Lz = sim.calculate_angular_momentum()
    L = np.sqrt(Lx**2 + Ly**2 + Lz**2)
    Lcirc = 0
    Mint = ps[0].m
    for p in ps[1:sim.N_real]: # the exact choice of which a and masses to use doesn't matter for closely packed systems (only hierarchical)
        mred = p.m*Mint/(p.m+Mint)
        Lcirc += mred * np.sqrt(sim.G*(p.m + Mint)*p.a)
        Mint += p.m
    return Lcirc - L



from subprocess import call
from collections import OrderedDict
import sys

datapath = '../training_data/'

datasets = 'all' # either a list of folders ([resonant, TTVsystems/Kepler-431]) or 'all' or 'ttv' to expand
runfunc = 'features'#'orbtseries'#'orbsummaryfeaturesxgb'

kwargs = OrderedDict()
kwargs['Norbits'] = 1e4
kwargs['Nout'] = 80
kwargs['trio'] = [[1,2,3]]

foldername = runfunc
for key, val in kwargs.items():
    if key == 'trio':
        foldername += 'trio'
    else:
        foldername += '{0}{1}'.format(key, val)

if datasets == 'all':
    datasets = ['random', 'resonant', 'TTVsystems/KOI-1576/', 'nonressystems/Kepler-431/']

for dataset in list(datasets):
    folder = datapath + dataset + '/'
    #call('rm -f ' + folder + 'labels.csv', shell=True)
    #call('rm -f ' + folder + 'massratios.csv', shell=True)
    #call('rm -f ' + folder + 'runstrings.csv', shell=True)
    print('removing ' + folder + foldername)
    call('rm -rf "' + folder + foldername + '"', shell=True)
    call('rm -rf "' + foldername + '"', shell=True)



import rebound
import unittest
from spock import FeatureClassifier, NbodyRegressor
from spock.feature_functions import get_tseries
from spock.simsetup import init_sim_parameters

def unstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-4, P=1)
    sim.add(m=1.e-4, P=1.3)
    sim.add(m=1.e-4, P=1.6)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim

def longstablesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-7, P=1)
    sim.add(m=1.e-7, P=2.1)
    sim.add(m=1.e-7, P=4.5)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim

def solarsystemsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.7e-7, a=0.39, e=0.21)
    sim.add(m=2.4e-6, a=0.72, e=0.007)
    sim.add(m=3.e-6, a=1, e=0.017)
    sim.add(m=3.2e-7, a=1.52, e=0.09)
    sim.add(m=1.e-3, a=5.2, e=0.049)
    sim.add(m=2.9e-4, a=9.54, e=0.055)
    sim.add(m=4.4e-5, a=19.2, e=0.047)
    sim.add(m=5.2e-5, a=30.1, e=0.009)
    for p in sim.particles[1:]:
        p.r = p.a*(p.m/3)**(1/3)
    sim.move_to_com()
    sim.collision='line'
    sim.integrator="whfast"
    sim.dt = 0.05
    return sim

def hyperbolicsim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-5, a=-1., e=1.2)
    sim.add(m=1.e-5, a=2.)
    sim.add(m=1.e-5, a=3.)
    return sim

def escapesim():
    sim = rebound.Simulation()
    sim.add(m=1.)
    sim.add(m=1.e-12, P=3.14, e=0.03, l=0.5)
    sim.add(m=1.e-12, P=4.396, e=0.03, l=4.8)
    sim.add(m=1.e-12, a=100, e=0.999)
    return sim

def rescale(sim, dscale, tscale, mscale):                                                                      
    simr = rebound.Simulation()
    vscale = dscale/tscale 
    simr.G *= mscale*tscale**2/dscale**3

    for p in sim.particles:
        simr.add(m=p.m/mscale, x=p.x/dscale, y=p.y/dscale, z=p.z/dscale, vx=p.vx/vscale, vy=p.vy/vscale, vz=p.vz/vscale, r=p.r/dscale)

    return simr

class TestClassifier(unittest.TestCase):
    def setUp(self):
        self.model = FeatureClassifier()
    
    def test_list(self):
        stable_target = [0, 0, 0, 0.7]
        stable = self.model.predict_stable([hyperbolicsim(), escapesim(), unstablesim(), longstablesim()])
        self.assertEqual(stable[0], 0)
        self.assertEqual(stable[1], 0)
        self.assertEqual(stable[2], 0)
        self.assertGreater(stable[3], 0.7) 
    
    def test_sim_unchanged(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        sim.integrate(1.2)
        x0 = sim.particles[1].x
        p1 = self.model.predict_stable(sim)
        self.assertEqual(sim.particles[1].x, x0)

    def test_repeat(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1.e-5, P=1.)
        sim.add(m=1.e-5, P=2.)
        sim.add(m=1.e-5, P=3.)
        p1 = self.model.predict_stable(sim)
        p2 = self.model.predict_stable(sim)
        self.assertEqual(p1, p2)

    def test_same_trajectory(self):
        sim = longstablesim()
        init_sim_parameters(sim)
        _, _ = get_tseries(sim, (1e4, 80, [[1,2,3]]))
        x1 = sim.particles[1].x

        sim = longstablesim()
        nbody = NbodyRegressor()
        nbody.predict_stable(sim, tmax=1e4, archive_filename='temp.bin', archive_interval=1.e4)
        sa = rebound.SimulationArchive('temp.bin')
        sim = sa[-1]
        x2 = sim.particles[1].x
        self.assertAlmostEqual(x1, x2, delta=1.e-5)
   
    # when chaotic realization matters, probs will vary more (eg t_inst=2e4)
    def test_galilean_transformation(self):
        sim = longstablesim()
        sim.move_to_com()
        p_com = self.model.predict_stable(sim)

        sim = longstablesim()
        for p in sim.particles:
            p.vx += 1000
        p_moving = self.model.predict_stable(sim)
        self.assertAlmostEqual(p_com, p_moving, delta=1.e-2)
   
    def test_rescale_distances(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim)

        sim = longstablesim()
        sim = rescale(sim, dscale=1e10, tscale=1, mscale=1)
        p1 = self.model.predict_stable(sim)
        self.assertAlmostEqual(p0, p1, delta=1.e-2)
    
    def test_rescale_times(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim)

        sim = longstablesim()
        sim = rescale(sim, dscale=1, tscale=1e10, mscale=1)
        p1 = self.model.predict_stable(sim)
        self.assertAlmostEqual(p0, p1, delta=1.e-1)

    def test_rescale_masses(self):
        sim = longstablesim()
        p0 = self.model.predict_stable(sim)

        sim = longstablesim()
        sim = rescale(sim, dscale=1, tscale=1, mscale=1e10)
        p1 = self.model.predict_stable(sim)
        self.assertAlmostEqual(p0, p1, delta=1.e-2)
    
    def test_hyperbolic(self):
        sim = hyperbolicsim()
        self.assertEqual(self.model.predict_stable(sim), 0)
    
    def test_escape(self):
        sim = escapesim()
        self.assertEqual(self.model.predict_stable(sim), 0)
    
    def test_unstable_in_short_integration(self):
        sim = unstablesim()
        self.assertEqual(self.model.predict_stable(sim), 0)
    
    def test_solarsystem(self):
        sim = solarsystemsim()
        self.assertGreater(self.model.predict_stable(sim), 0.7)
    
    def test_stable(self):
        sim = longstablesim()
        self.assertGreater(self.model.predict_stable(sim), 0.7)
    
if __name__ == '__main__':
    unittest.main()



from runfunctions import run_resonant
import sys

seed = int(sys.argv[1])
runstring = sys.argv[2]
run_resonant(seed, runstr=runstring) # run system
run_resonant(seed, shadow=True, runstr=runstring) # run shadow system



import rebound
import numpy as np
import pandas as pd
from collections import OrderedDict
from spock.feature_functions import get_pairs, find_strongest_MMR, populate_trio
from spock.AMD_functions import AMD, AMD_crit

def additional_get_tseries(sim, args):
    Norbits = args[0]
    Nout = args[1]
    trios = args[2]

    P0 = sim.particles[1].P
    times = np.linspace(0, Norbits*P0, Nout)
    
    triopairs, triotseries = [], []
    for tr, trio in enumerate(trios): # For each trio there are two adjacent pairs 
        triopairs.append(get_pairs(sim, trio))
        triotseries.append(np.zeros((Nout, 9))*np.nan)
   
    for i, time in enumerate(times):
        try:
            sim.integrate(time, exact_finish_time=0)
        except:
            pass

        if sim._status == 5: # checking this way works for both new rebound and old version used for random dataset
            stable = False
            return triotseries, stable

        for tseries in triotseries:
            tseries[i,0] = sim.t/P0  # time

        for tr, trio in enumerate(trios):
            pairs = triopairs[tr]
            tseries = triotseries[tr] 
            populate_trio(sim, trio, pairs, tseries, i)
            tseries[i,8] = AMD(sim)

    stable = True
    return triotseries, stable
    
def additional_features(sim, args): # final cut down list
    Norbits = args[0]
    Nout = args[1]
    trios = args[2]
    
    ps  = sim.particles
    triofeatures = []
    for tr, trio in enumerate(trios):
        features = OrderedDict()
        pairs = get_pairs(sim, trio)
        for i, [label, i1, i2] in enumerate(pairs):
            features['EMfracstd'+label] = np.nan
            features['EPstd'+label] = np.nan
            features['AMDfrac'+label] = np.nan
            features['MMRstrength'+label] = np.nan
            
            RH = ps[i1].a*((ps[i1].m + ps[i2].m)/ps[0].m)**(1./3.)
            features['beta'+label] = (ps[i2].a-ps[i1].a)/RH
            features["AMDcrit"+label] = AMD_crit(sim, i1, i2)
            features["EMcross"+label] = (ps[i2].a-ps[i1].a)/ps[i1].a       
            features["j"+label], features["k"+label], _ = find_strongest_MMR(sim, i1, i2) 

        features['MEGNO'] = np.nan
        features['MEGNOstd'] = np.nan
        features['stable_in_short_integration'] = False

        triofeatures.append(features)
    
    features["e1Z07"] = ps[1].e * (ps[2].a+ps[1].a) / (ps[2].a-ps[1].a)
    features["e2Z07"] = ps[2].e * (ps[3].a+ps[2].a) / (ps[3].a-ps[2].a)
    features["e3Z07"] = ps[3].e * (ps[3].a+ps[2].a) / (ps[3].a-ps[2].a)
    features["eavgZ07inner"] = np.mean([features['e1Z07'], features['e2Z07']])
    features["eavgZ07outer"] = np.mean([features['e2Z07'], features['e3Z07']])
    features["eavgZ07"] = np.mean([features["e1Z07"], features["e2Z07"], features["e3Z07"]])
    features["muavgZ07inner"] = (ps[1].m+ps[2].m)/ps[0].m/2 # mean of the mass ratios
    features["muavgZ07outer"] = (ps[2].m+ps[3].m)/ps[0].m/2 # mean of the mass ratios
    features["muavgZ07"] = np.mean([ps[1].m, ps[2].m, ps[3].m])
    features["kZ07inner"] = (ps[2].a-ps[1].a) * 2. / (ps[2].a + ps[1].a) / (2.*features["muavgZ07inner"]/3.)**(1./3.)
    features["kZ07outer"] = (ps[3].a-ps[2].a) * 2. / (ps[3].a + ps[2].a) / (2.*features["muavgZ07outer"]/3.)**(1./3.)
    features["kZ07avg"] = np.mean([features['kZ07inner'], features['kZ07outer']])
    features["AZ07inner"] = -2. + features["eavgZ07inner"] - 0.27*np.log10(features["muavgZ07inner"]) # Zhou 2007 Eq 4
    features["AZ07outer"] = -2. + features["eavgZ07outer"] - 0.27*np.log10(features["muavgZ07outer"]) # Zhou 2007 Eq 4
    features["AZ07avg"] = -2. + features["eavgZ07"] - 0.27*np.log10(features["muavgZ07"]) # Zhou 2007 Eq 4
    features["BZ07inner"] = 18.7 + 1.1*np.log10(features["muavgZ07inner"]) - (16.8 + 1.2*np.log10(features["muavgZ07inner"]))*features['eavgZ07inner'] # Zhou 2007 Eq 4
    features["BZ07outer"] = 18.7 + 1.1*np.log10(features["muavgZ07outer"]) - (16.8 + 1.2*np.log10(features["muavgZ07outer"]))*features['eavgZ07outer'] # Zhou 2007 Eq 4
    features["BZ07avg"] = 18.7 + 1.1*np.log10(features["muavgZ07"]) - (16.8 + 1.2*np.log10(features["muavgZ07"]))*features['eavgZ07'] # Zhou 2007 Eq 4
    features["Z07log_instability_time_inner"] = features["AZ07inner"] + features['BZ07inner']*np.log10(features['kZ07inner']/2.3)
    features["Z07log_instability_time_outer"] = features["AZ07outer"] + features['BZ07outer']*np.log10(features['kZ07outer']/2.3)
    features["Z07log_instability_time_avg"] = features["AZ07avg"] + features['BZ07avg']*np.log10(features['kZ07avg']/2.3)
    features["Z07Stable_avg"] = features["Z07log_instability_time_avg"] > 9
    features["Z07Stable_worstpair"] = min(features["Z07log_instability_time_inner"], features["Z07log_instability_time_outer"]) > 9

    features["deltaQ11inner"] = (ps[2].a-ps[1].a)/ps[2].a
    features["deltaQ11outer"] = (ps[3].a-ps[2].a)/ps[3].a
    features["deltaQ11avg"] = np.mean([features['deltaQ11inner'], features['deltaQ11outer']])
    features["Q11log_instability_time_inner"] = np.log10(features["deltaQ11inner"]**8 / np.abs(np.log(features["deltaQ11inner"]))**3 / features["muavgZ07inner"]**3 / 8.) # Qullen 2011 Eq 68
    features["Q11log_instability_time_outer"] = np.log10(features["deltaQ11outer"]**8 / np.abs(np.log(features["deltaQ11outer"]))**3 / features["muavgZ07outer"]**3 / 8.) # Qullen 2011 Eq 68
    features["Q11log_instability_time_avg"] = np.log10(features["deltaQ11avg"]**8 / np.abs(np.log(features["deltaQ11avg"]))**3 / features["muavgZ07"]**3 / 8.) # Qullen 2011 Eq 68
    features["Q11Stable_avg"] = features['Q11log_instability_time_avg'] > 9
    features["Q11Stable_worstpair"] = min(features['Q11log_instability_time_inner'], features["Q11log_instability_time_outer"]) > 9

    triotseries, stable = additional_get_tseries(sim, args)
    if not stable:
        return triofeatures, stable

    for features, tseries in zip(triofeatures, triotseries):
        EMnear = tseries[:, 1]
        EPnear = tseries[:, 2]
        MMRstrengthnear = tseries[:,3]
        EMfar = tseries[:, 4]
        EPfar = tseries[:, 5]
        MMRstrengthfar = tseries[:,6]
        MEGNO = tseries[:, 7]
        AMD = tseries[:, 8]

        features['MEGNO'] = np.median(MEGNO[-int(Nout/10):]) # smooth last 10% to remove oscillations around 2
        features['MEGNOstd'] = MEGNO[int(Nout/5):].std()
        features['AMDfracnear'] = np.median(AMD) / features['AMDcritnear']
        features['AMDfracfar'] = np.median(AMD) / features['AMDcritfar']
        features['MMRstrengthnear'] = np.median(MMRstrengthnear)
        features['MMRstrengthfar'] = np.median(MMRstrengthfar)
        features['EMfracstdnear'] = EMnear.std() / features['EMcrossnear']
        features['EMfracstdfar'] = EMfar.std() / features['EMcrossfar']
        features['EPstdnear'] = EPnear.std() 
        features['EPstdfar'] = EPfar.std() 
        
    return triofeatures, stable



import rebound
from multiprocessing import Pool
from itertools import repeat
from runfunctions import run_random
import pandas as pd
import sys
import math

N_systems=2000
tmax = 1.e6
betamin=5
betamax=20

integrator=sys.argv[1]
dt = float(sys.argv[2])*math.sqrt(3)/100.
shadow = int(sys.argv[3])
Nthreads = int(sys.argv[4])
print(dt)

params = list(zip(range(N_systems), repeat(integrator), repeat(dt), repeat(tmax), repeat(betamin), repeat(betamax), repeat(shadow)))

pool = Pool(Nthreads)
res = pool.starmap(run_random, params)
df = pd.DataFrame(res, columns=('instability_time', 'frac_energy_error', 'wall_time'))

datapath = '../data/whfastias15comparison/'
if integrator=="ias15":
    filename = "ias15"
else:
    filename = "whfastdt{:.3f}".format(dt)
if shadow == 1:
    filename += "shadow"
filename += ".csv"

df.to_csv(datapath+filename, encoding='ascii')



# Run script with python run.py sim_id, where sim_id is an integer to use for the sim_id (and the RNG seed). Runs a base system and a shadow system
import numpy as np
import rebound
from random import random, uniform, seed
import time
import sys
import math
import random
from celmech import Andoyer, Poincare, AndoyerHamiltonian
from celmech.andoyer import get_Xstarres
from scipy.integrate import ode

def logunif(r, mini, maxi):
    logmin = np.log10(mini)
    logmax = np.log10(maxi)
    return 10**(r.uniform(logmin, logmax))

def get_resonant(seed, Nplanets=3):
    r = random.Random()
    r.seed(seed)

    Phiprimecrits = [0, 1., -2./3.]
    pairs = ['inner', 'outer', 'split']
    
    k=r.randint(1,2) 
    pairindex=r.randint(0,2)
    pair = pairs[pairindex]
    m1 = logunif(r, 1.e-7, 1.e-4)
    m2 = logunif(r, 1.e-7, 1.e-4)
    eH = ((m1+m2)/3.)**(1./3.)
    ehillstable = 3.5*eH
    jmax = k/(1-1/(1+3.5*eH)**1.5)
    if pair == 'split':
        if Nplanets == 2:
            return # don't want 2planet systems 60 hill radii apart
        maxHillradii = 60. # 3rd planet will go in middle so draw up to 60
    else:
        maxHillradii = 30.
    jmin = max(k+1, k/(1-1/(1+maxHillradii*eH)**1.5))
    jmin = int(np.ceil(jmin))
    jmax = int(np.floor(jmax))
    if k == 2: # if k = 2, want odd j so we don't get e.g. 8:6 = 4:3
        if jmin % 2 == 0: # even
            jmin += 1
        if jmax % 2 == 0:
            jmax -= 1
    j = r.randrange(jmin, jmax+1, k) # choose randomly between limits in steps of k e.g. (3,5,7,9)
    a1 = 1.
    a2 = (float(j)/(j-k))**(2./3.)
    ecross1 = (a2-a1)/a1
    ecross2 = (a2-a1)/a2
    emin1 = m2/ecross1**2
    emin2 = m1/ecross2**2
    emin = max(emin1, emin2) # take as min Z the larger of the kicks a planet gets at conjunction       
    emin = max(emin, (m1+m2)**(1./k)) # below mtot^1/k, the resonant term is smaller than the second order mass terms we ignore
    emax = min(ecross1, ecross2)
    
    avars = Andoyer(j=j,k=k,X=0,Y=0,m1=m1, m2=m2)
    Phiprimecrit = Phiprimecrits[k]
    Xcrit = get_Xstarres(k, Phiprimecrit)
    Phicrit = 0.5*Xcrit**2
    emin = max(avars.Phi_to_Z(Phicrit), emin) # first quantity is value of Z at bifurcation when res first appears
    
    Zstar=logunif(r, emin, emax)
    libfac=logunif(r, 3.e-3, 3)
    negative = r.randint(0,1)
    if negative:
        libfac *= -1
   
    Zcom = logunif(r, emin, emax)
    avars = Andoyer.from_elements(j=j,k=k,Zstar=Zstar,libfac=libfac, m1=m1, m2=m2, Zcom=Zcom, phiZcom=r.uniform(0,2*np.pi), theta=r.uniform(0,2*np.pi), theta1=r.uniform(0,2*np.pi))
    tmax = r.uniform(0,10*avars.tlib)
    H = AndoyerHamiltonian(avars)
    H.integrate(tmax)
    pvars = avars.to_Poincare()
    ps = pvars.particles
    if Nplanets==3:
        m3 = logunif(r, 1.e-7, 1.e-4)
        pvarssorted = Poincare(G=pvars.G)

        if pair == "inner":
            eH = ((m2+m3)/3.)**(1./3.)
            beta = r.uniform(3.5,30)
            a3 = a2*(1+beta*eH)
            ecross3 = (a3-a2)/a3
            emin3 = m2/ecross3**2
            e3 = logunif(r, emin3, ecross3)
            pvarssorted.add(m=ps[1].m, M=ps[1].M, a=ps[1].a, e=ps[1].e, gamma=ps[1].gamma, l=ps[1].l)
            pvarssorted.add(m=ps[2].m, M=ps[2].M, a=ps[2].a, e=ps[2].e, gamma=ps[2].gamma, l=ps[2].l)
            pvarssorted.add(m=m3, M=1, a=a3, e=e3, gamma=r.uniform(0, 2*np.pi), l=r.uniform(0, 2*np.pi))

        elif pair == "outer":
            eH = ((m1+m3)/3.)**(1./3.)
            beta = r.uniform(3.5,30)
            a3 = a1/(1+beta*eH)
            ecross3 = (a1-a3)/a3
            emin3 = m1/ecross3**2
            e3 = logunif(r, emin3, ecross3)
            pvarssorted.add(m=m3, M=1, a=a3, e=e3, gamma=r.uniform(0, 2*np.pi), l=r.uniform(0, 2*np.pi))
            pvarssorted.add(m=ps[1].m, M=ps[1].M, a=ps[1].a, e=ps[1].e, gamma=ps[1].gamma, l=ps[1].l)
            pvarssorted.add(m=ps[2].m, M=ps[2].M, a=ps[2].a, e=ps[2].e, gamma=ps[2].gamma, l=ps[2].l)

        elif pair == "split":
            eH1 = ((m1+m3)/3.)**(1./3.)
            eH2 = ((m2+m3)/3.)**(1./3.)
            amin = a1*(1+3.5*eH1)
            amax = min(a1*(1+30*eH1), a2/(1+3.5*eH2))
            if amin > amax:
                seed += 900000
                return get_resonant(seed=seed) # draw new sample (may not be 'split')
            a3 = r.uniform(amin, amax)
            ecross3 = min((a3-a1)/a1, (a2-a3)/a3)
            emin3 = max(m1/((a3-a1)/a3)**2, m2/((a2-a3)/a3)**2)
            e3 = logunif(r, emin3, ecross3)
            pvarssorted.add(m=ps[1].m, M=ps[1].M, a=ps[1].a, e=ps[1].e, gamma=ps[1].gamma, l=ps[1].l)
            pvarssorted.add(m=m3, M=1, a=a3, e=e3, gamma=r.uniform(0, 2*np.pi), l=r.uniform(0, 2*np.pi))
            pvarssorted.add(m=ps[2].m, M=ps[2].M, a=ps[2].a, e=ps[2].e, gamma=ps[2].gamma, l=ps[2].l)
        
        sim = pvarssorted.to_Simulation()
    else:
        sim = avars.to_Simulation()
    # add inclinations and scale s.t. a1 = 1, Mprimary=1. and G = 4*pi**2
    ps = sim.particles
    
    siminc = rebound.Simulation()
    siminc.G = 4*np.pi**2
    
    dscale = ps[1].a
    tscale = ps[1].P
    mscale = siminc.G*dscale**3/tscale**2 # ps[0].m + ps[1].m
    
    siminc.add(m=ps[0].m/mscale, x=ps[0].x/dscale, y=ps[0].y/dscale, vx=ps[0].vx/dscale*tscale, vy=ps[0].vy/dscale*tscale)
    for p in ps[1:]:
        siminc.add(m=p.m/mscale, a=p.a/dscale, e=p.e, inc=logunif(r, 1.e-3, 1.e-1), Omega=r.uniform(0,2*np.pi), pomega=p.pomega, l=p.l)
        rH = siminc.particles[-1].a*(siminc.particles[-1].m/3./siminc.particles[0].m)**(1./3.) 
        siminc.particles[-1].r = rH
    siminc.move_to_com()
    siminc.integrator="whfast"
    siminc.dt = 2.*np.sqrt(3)/100.*siminc.particles[1].P
    siminc.ri_whfast.safe_mode=0
    siminc.collision="line"
    return siminc, j, k, pairindex, Zstar, libfac, Zcom

def run_resonant(seed, runstr, Nplanets=3, maxorbs=1.e9, shadow=False):
    originalseed = seed
    success=False
    while not success:
        try:
            sim, j, k, pairindex, Zstar, libfac, Zcom = get_resonant(seed, Nplanets)
            sim.integrate(1.e4*sim.particles[1].P)
            success=True
        except rebound.Collision:
            seed += 700000 # roughly relatively prime with amin > amax offset above
    sim, j, k, pairindex, Zstar, libfac, Zcom = get_resonant(seed, Nplanets) # get same sim
    print('{0}, {1}, {2}, {3}, {4}, {5:8e}, {6:8e}, {7:8e}\n'.format(originalseed, seed, pairindex, j, k, Zstar, libfac, Zcom))
    if shadow:
        shadowstr = 'shadow'
        kicksize=1.e-11
        sim.particles[2].x += kicksize
    else:
        shadowstr = ''
    filename = '../data/resonant/simulation_archives/'+shadowstr+'runs/sa'+runstr
    sim.automateSimulationArchive(filename, interval=maxorbs/1000., deletefile=True)
    try:
        sim.integrate(maxorbs*sim.particles[1].P)
    except rebound.Collision:
        sim.simulationarchive_snapshot(filename)  # save final snapshot if collision occurs

def run_resonant_condition(seed, runstr, conditionfunc, Nplanets=3, maxorbs=1.e9, shadow=False):
    originalseed = seed
    sim, j, k, pairindex, Zstar, libfac, Zcom = get_resonant(seed, Nplanets)
    if conditionfunc(sim):
        if shadow:
            shadowstr = 'shadow'
            kicksize=1.e-11
            sim.particles[2].x += kicksize
        else:
            shadowstr = ''
        filename = '../data/resonant/simulation_archives/'+shadowstr+'runs/sa'+runstr
        sim.automateSimulationArchive(filename, interval=maxorbs/1000., deletefile=True)
        try:
            sim.integrate(maxorbs*sim.particles[1].P)
        except rebound.Collision:
            sim.simulationarchive_snapshot(filename)  # save final snapshot if collision occurs

def collision(reb_sim, col):
    reb_sim.contents._status = 5 # causes simulation to stop running and have flag for whether sim stopped due to collision
    return 0

def run_random(sim_id, integrator="whfast", dt=None, maxorbs=1.e9, betamin=1., betamax=30., shadow=False, runstr=None):
    if dt is None:
        dt = 2.*math.sqrt(3)/100. # dt approx 3.5 % of innermost orbital period

    datapath = '../data/random/'

    a1 = 1. # All distances in units of the innermost semimajor axis (always at 1)
    Mstar = 1. # All masses in units of stellar mass

    logMmin = np.log10(1.e-7) # 1/3 Mars around Sun
    logMmax = np.log10(1.e-4) # 2 Nep around Sun
    logincmin = np.log10(1.e-3)
    logincmax = np.log10(1.e-1) # max mutual inclination of 11.4 degrees. Absolute of 5.7 deg

    seed(sim_id)

    M1 = 10.**uniform(logMmin, logMmax)
    M2 = 10.**uniform(logMmin, logMmax)
    M3 = 10.**uniform(logMmin, logMmax)

    hill12 = a1*((M1+M2)/3.)**(1./3.)
    beta1 = uniform(betamin, betamax)
    a2 = a1 + beta1*hill12

    hill23 = a2*((M2+M3)/3.)**(1./3.)
    beta2 = uniform(betamin, betamax)
    a3 = a2 + beta2*hill23

    minhill = min(hill12, hill23)

    ecrit1 = (a2-a1)/a1
    ecrit21 = (a2-a1)/a2
    ecrit23 = (a3-a2)/a2
    ecrit3 = (a3-a2)/a3
    
    logemax1 = np.log10(ecrit1)
    logemax2 = np.log10(min(ecrit21, ecrit23))
    logemax3 = np.log10(ecrit3)

    logemin1 = np.log10(M2/ecrit1**2)
    logemin2 = np.log10(max(M1/ecrit21**2, M3/ecrit23**2))
    logemin3 = np.log10(M2/ecrit3**2)

    #print("M1 = {0}, M2 = {1}, M3 = {2}".format(M1, M2, M3))
    #print("beta1 = {0}, beta2 = {1}".format(beta1, beta2))
    #print("a1 = {0}, a2 = {1}, a3 = {2}".format(a1, a2, a3))
    #print("emax1 = {0}, emax2 = {1}, emax3 = {2}".format(10**logemax1, 10**logemax2, 10**logemax3))
    #print("emin1 = {0}, emin2 = {1}, emin3 = {2}".format(10**logemin1, 10**logemin2, 10**logemin3))
    #print("minhill = {0}".format(minhill))

    emax = 0.999
    e1 = min(10.**uniform(logemin1, logemax1), emax) # make sure ecc < 1
    e2 = min(10.**uniform(logemin2, logemax2), emax)
    e3 = min(10.**uniform(logemin3, logemax3), emax)
    
    i1 = 10.**uniform(logincmin, logincmax)
    i2 = 10.**uniform(logincmin, logincmax)
    i3 = 10.**uniform(logincmin, logincmax)

    sim = rebound.Simulation()
    sim.integrator=integrator
    sim.ri_whfast.safe_mode = 0
    sim.G = 4*np.pi**2

    sim.add(m=1.)
    sim.add(m=M1, a=a1, e=e1, pomega=random()*2.*np.pi, inc=i1, Omega=random()*2.*np.pi, f=random()*2.*np.pi, r=minhill)
    sim.add(m=M2, a=a2, e=e2, pomega=random()*2.*np.pi, inc=i2, Omega=random()*2.*np.pi, f=random()*2.*np.pi, r=minhill)
    sim.add(m=M3, a=a3, e=e3, pomega=random()*2.*np.pi, inc=i3, Omega=random()*2.*np.pi, f=random()*2.*np.pi, r=minhill)
    sim.move_to_com()
    ps = sim.particles

    if shadow:
        kicksize=1.e-11
        ps[2].x += kicksize

    if integrator=="whfast":
        sim.dt = dt*sim.particles[1].P
    sim.collision = "direct"
    sim.collision_resolve = collision
        
    #runstr = "{0:0=7d}.bin".format(sim_id)
    if shadow:
        shadowstr = 'shadow'
    else:
        shadowstr = ''

    if runstr:
        sim.save(datapath+'initial_conditions/'+shadowstr+'runs/ic'+runstr)
        sim.initSimulationArchive(datapath+'simulation_archives/'+shadowstr+'runs/sa'+runstr, interval=maxorbs/1000.)

    E0 = sim.calculate_energy()
    t0 = time.time()
    sim.integrate(maxorbs) # will stop if collision occurs
    Ef = sim.calculate_energy()    
    Eerr = abs((Ef-E0)/E0)

    if runstr:
        sim.save(datapath+'final_conditions/'+shadowstr+'runs/fc'+runstr)

    return (sim.t, Eerr, time.time()-t0)



import pickle as pkl
from copy import deepcopy as copy
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, PowerTransformer
import matplotlib as mpl
mpl.use('agg')
import numpy as np
from matplotlib import pyplot as plt
import torch
from torch import nn
from torch.autograd import Variable
import sys
import torch.nn.functional as F
from torch.nn import Parameter
import math
import pytorch_lightning as pl
from pytorch_lightning import Trainer
import math
from torch._six import inf
from functools import wraps
import warnings
from torch.optim.optimizer import Optimizer
from collections import OrderedDict


class CustomOneCycleLR(torch.optim.lr_scheduler._LRScheduler):
    """Custom version of one-cycle learning rate to stop early"""
    def __init__(self,
                 optimizer,
                 max_lr,
                 swa_steps_start,
                 pct_start=0.3,
                 anneal_strategy='cos',
                 cycle_momentum=True,
                 base_momentum=0.85,
                 max_momentum=0.95,
                 div_factor=25.,
                 final_div_factor=1e4,
                 last_epoch=-1):

        total_steps = swa_steps_start #Just fix afterwards.
        epochs = None
        steps_per_epoch = None
        # Validate optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        # Validate total_steps
        if total_steps is None and epochs is None and steps_per_epoch is None:
            raise ValueError("You must define either total_steps OR (epochs AND steps_per_epoch)")
        elif total_steps is not None:
            if total_steps <= 0 or not isinstance(total_steps, int):
                raise ValueError("Expected non-negative integer total_steps, but got {}".format(total_steps))
            self.total_steps = total_steps
        else:
            if epochs <= 0 or not isinstance(epochs, int):
                raise ValueError("Expected non-negative integer epochs, but got {}".format(epochs))
            if steps_per_epoch <= 0 or not isinstance(steps_per_epoch, int):
                raise ValueError("Expected non-negative integer steps_per_epoch, but got {}".format(steps_per_epoch))
            self.total_steps = epochs * steps_per_epoch
        self.step_size_up = float(pct_start * self.total_steps) - 1
        self.step_size_down = float(self.total_steps - self.step_size_up) - 1

        # Validate pct_start
        if pct_start < 0 or pct_start > 1 or not isinstance(pct_start, float):
            raise ValueError("Expected float between 0 and 1 pct_start, but got {}".format(pct_start))

        # Validate anneal_strategy
        if anneal_strategy not in ['cos', 'linear']:
            raise ValueError("anneal_strategy must by one of 'cos' or 'linear', instead got {}".format(anneal_strategy))
        elif anneal_strategy == 'cos':
            self.anneal_func = self._annealing_cos
        elif anneal_strategy == 'linear':
            self.anneal_func = self._annealing_linear

        # Initialize learning rate variables
        max_lrs = self._format_param('max_lr', self.optimizer, max_lr)
        if last_epoch == -1:
            for idx, group in enumerate(self.optimizer.param_groups):
                group['initial_lr'] = max_lrs[idx] / div_factor
                group['max_lr'] = max_lrs[idx]
                group['min_lr'] = group['initial_lr'] / final_div_factor

        # Initialize momentum variables
        self.cycle_momentum = cycle_momentum
        if self.cycle_momentum:
            if 'momentum' not in self.optimizer.defaults and 'betas' not in self.optimizer.defaults:
                raise ValueError('optimizer must support momentum with `cycle_momentum` option enabled')
            self.use_beta1 = 'betas' in self.optimizer.defaults
            max_momentums = self._format_param('max_momentum', optimizer, max_momentum)
            base_momentums = self._format_param('base_momentum', optimizer, base_momentum)
            if last_epoch == -1:
                for m_momentum, b_momentum, group in zip(max_momentums, base_momentums, optimizer.param_groups):
                    if self.use_beta1:
                        _, beta2 = group['betas']
                        group['betas'] = (m_momentum, beta2)
                    else:
                        group['momentum'] = m_momentum
                    group['max_momentum'] = m_momentum
                    group['base_momentum'] = b_momentum

        super(CustomOneCycleLR, self).__init__(optimizer, last_epoch)

    def _format_param(self, name, optimizer, param):
        """Return correctly formatted lr/momentum for each param group."""
        if isinstance(param, (list, tuple)):
            if len(param) != len(optimizer.param_groups):
                raise ValueError("expected {} values for {}, got {}".format(
                    len(optimizer.param_groups), name, len(param)))
            return param
        else:
            return [param] * len(optimizer.param_groups)

    def _annealing_cos(self, start, end, pct):
        "Cosine anneal from `start` to `end` as pct goes from 0.0 to 1.0."
        if pct >= 1.0:
            return end
        cos_out = math.cos(math.pi * pct) + 1
        return end + (start - end) / 2.0 * cos_out

    def _annealing_linear(self, start, end, pct):
        "Linearly anneal from `start` to `end` as pct goes from 0.0 to 1.0."
        if pct >= 1.0:
            return end
        return (end - start) * pct + start

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", DeprecationWarning)

        lrs = []
        step_num = self.last_epoch
        if step_num > self.total_steps:
            raise ValueError("Tried to step {} times. The specified number of total steps is {}"
                             .format(step_num + 1, self.total_steps))
        for group in self.optimizer.param_groups:
            if step_num <= self.step_size_up:
                computed_lr = self.anneal_func(group['initial_lr'], group['max_lr'], step_num / self.step_size_up)
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(group['max_momentum'], group['base_momentum'],
                                                         step_num / self.step_size_up)
            else:
                down_step_num = step_num - self.step_size_up
                computed_lr = self.anneal_func(group['max_lr'], group['min_lr'], down_step_num / self.step_size_down)
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(group['base_momentum'], group['max_momentum'],
                                                         down_step_num / self.step_size_down)
            lrs.append(computed_lr)
            if self.cycle_momentum:
                if self.use_beta1:
                    _, beta2 = group['betas']
                    group['betas'] = (computed_momentum, beta2)
                else:
                    group['momentum'] = computed_momentum
        return lrs

def get_data(
        ssX=None,
        batch_size=32,
        train=True,
        **kwargs):
    """
    inputs:
        batch_size: int

    return:
        (dataloader, test_dataloader)
    """
    plot_random = False if 'plot_random' not in kwargs else kwargs['plot_random']
    plot_resonant = not plot_random
    train_all = False if 'train_all' not in kwargs else kwargs['train_all']
    plot = False if 'plot' not in kwargs else kwargs['plot']
    if not train_all and ssX is None:
        plot_resonant = True
        plot_random = False

    if train_all:
        filename = 'data/combined.pkl'
    elif plot_resonant:
        filename = 'data/resonant_dataset.pkl'
    elif plot_random:
        filename = 'data/random_dataset.pkl'

    # These are generated by data_from_pkl.py
    loaded_data = pkl.load(
        open(filename, 'rb')
    )

    train_ssX = (ssX is None)

    fullX, fully = loaded_data['X'], loaded_data['y']

    if train_all:
        len_random = 17082 #Number of valid random examples (others have NaNs)
        random_data = np.arange(len(fullX)) >= (len(fullX) - len_random)


    # Differentiate megno
    if 'fix_megno' in kwargs and kwargs['fix_megno']:
        idx = [i for i, lab in enumerate(loaded_data['labels']) if 'megno' in lab][0]
        fullX[:, 1:, idx] -= fullX[:, :-1, idx]

    if 'include_derivatives' in kwargs and kwargs['include_derivatives']:
        derivative = fullX[:, 1:, :] - fullX[:, :-1, :]
        derivative = np.concatenate((
            derivative[:, [0], :],
            derivative), axis=1)
        fullX = np.concatenate((
            fullX, derivative),
            axis=2)


    # Hide fraction of test
    # MAKE SURE WE DO COPIES AFTER!!!!
    if train:
        if train_all:
            remy, finaly, remX, finalX, rem_random, final_random = train_test_split(fully, fullX, random_data, shuffle=True, test_size=1./10, random_state=0)
            trainy, testy, trainX, testX, train_random, test_random = train_test_split(remy, remX, rem_random, shuffle=True, test_size=1./10, random_state=1)
        else:
            remy, finaly, remX, finalX = train_test_split(fully, fullX, shuffle=True, test_size=1./10, random_state=0)
            trainy, testy, trainX, testX = train_test_split(remy, remX, shuffle=True, test_size=1./10, random_state=1)
    else:
        assert not train_all
        remy = fully
        finaly = fully
        testy = fully
        trainy = fully
        remX = fullX
        finalX = fullX
        testX = fullX
        trainX = fullX

    if plot:
        # Use test dataset for plotting, so put it in validation part:
        testX = finalX
        testy = finaly

    if train_ssX:
        if 'power_transform' in kwargs and kwargs['power_transform']:
            ssX = PowerTransformer(method='yeo-johnson') #Power is best
        else:
            ssX = StandardScaler() #Power is best

    n_t = trainX.shape[1]
    n_features = trainX.shape[2]

    if train_ssX:
        ssX.fit(trainX.reshape(-1, n_features)[::1539])

    ttrainy = trainy
    ttesty = testy
    ttrainX = ssX.transform(trainX.reshape(-1, n_features)).reshape(-1, n_t, n_features)
    ttestX = ssX.transform(testX.reshape(-1, n_features)).reshape(-1, n_t, n_features)
    if train_all:
        ttest_random = test_random
        ttrain_random = train_random

    tremX = ssX.transform(remX.reshape(-1, n_features)).reshape(-1, n_t, n_features)
    tremy = remy

    train_len = ttrainX.shape[0]
    X = Variable(torch.from_numpy(np.concatenate((ttrainX, ttestX))).type(torch.FloatTensor))
    y = Variable(torch.from_numpy(np.concatenate((ttrainy, ttesty))).type(torch.FloatTensor))
    if train_all:
        r = Variable(torch.from_numpy(np.concatenate((ttrain_random, ttest_random))).type(torch.BoolTensor))

    Xrem = Variable(torch.from_numpy(tremX).type(torch.FloatTensor))
    yrem = Variable(torch.from_numpy(tremy).type(torch.FloatTensor))

    idxes = np.s_[:]
    dataset = torch.utils.data.TensorDataset(X[:train_len, :, idxes], y[:train_len])
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=8)

    # Cut up dataset into only the random or resonant parts. 
    # Only needed if plotting OR 
    if (not plot) or (not train_all):
        test_dataset = torch.utils.data.TensorDataset(X[train_len:, :, idxes], y[train_len:])
    else:
        if plot_random: mask =  r
        else:           mask = ~r
        print(f'Plotting with {mask.sum()} total elements, when plot_random={plot_random}')
        test_dataset = torch.utils.data.TensorDataset(X[train_len:][r[train_len:]][:, :, idxes], y[train_len:][r[train_len:]])

    test_dataloader = torch.utils.data.DataLoader(test_dataset, batch_size=3000, shuffle=False, pin_memory=True, num_workers=8)
        
    kwargs['model'].ssX = copy(ssX)

    return dataloader, test_dataloader


def soft_clamp(x, lo, high):
    return 0.5*(torch.tanh(x)+1)*(high-lo) + lo

Linear = nn.Linear
module = nn.Module

def mlp(in_n, out_n, hidden, layers, act='relu'):
    if act == 'relu':
        act = nn.ReLU
    elif act == 'softplus':
        act = nn.Softplus
    else:
        raise NotImplementedError('act must be relu or softplus')

    if layers == 0:
        return Linear(in_n, out_n)

    result = [Linear(in_n, hidden),
             act()]
    for i in reversed(range(layers)):
        result.extend([
            Linear(hidden, hidden),
            act()
            ])

    result.extend([nn.Linear(hidden, out_n)])
    return nn.Sequential(*result)

def safe_log_erf(x):
    base_mask = x < -1
    value_giving_zero = torch.zeros_like(x, device=x.device)
    x_under = torch.where(base_mask, x, value_giving_zero)
    x_over = torch.where(~base_mask, x, value_giving_zero)
    
    f_under = lambda x: (
         0.485660082730562*x + 0.643278438654541*torch.exp(x) + 
         0.00200084619923262*x**3 - 0.643250926022749 - 0.955350621183745*x**2
    )
    f_over = lambda x: torch.log(1.0+torch.erf(x))
    
    return f_under(x_under) + f_over(x_over)

EPSILON = 1e-5

class VarModel(pl.LightningModule):
    """Bayesian Neural Network model for predicting instability time"""
    def __init__(self, hparams):
        super().__init__()
        if 'seed' not in hparams: hparams['seed'] = 0
        # pl.seed_everything(hparams['seed'])

        hparams['include_derivatives'] = False if 'include_derivatives' not in hparams else hparams['include_derivatives']

        if 'time_series_features' not in hparams:
            hparams['time_series_features'] = 38+3

        if hparams['time_series_features'] == 82:
            hparams['time_series_features'] = 41

        self.fix_megno = False if 'fix_megno' not in hparams else hparams['fix_megno']
        self.fix_megno2 = False if 'fix_megno2' not in hparams else hparams['fix_megno2']
        self.include_angles = False if 'include_angles' not in hparams else hparams['include_angles']

        self.n_features = hparams['time_series_features'] * (1 + int(hparams['include_derivatives']))
        self.feature_nn = mlp(self.n_features, hparams['latent'], hparams['hidden'], hparams['in'])
        self.regress_nn = mlp(hparams['latent']*2 + int(self.fix_megno)*2, 2, hparams['hidden'], hparams['out'])
        self.input_noise_logvar = nn.Parameter(torch.zeros(self.n_features)-2)
        self.summary_noise_logvar = nn.Parameter(torch.zeros(hparams['latent'] * 2 + int(self.fix_megno)*2) - 2) # add to summaries, not direct latents
        self.lowest = 0.5
        if 'lower_std' in hparams and hparams['lower_std']:
            self.lowest = 0.1

        self.latents = None
        self.beta_in = 1 if 'beta_in' not in hparams else hparams['beta_in']
        self.beta_out = 1 if 'beta_out' not in hparams else hparams['beta_out']
        self.megno_location = 7
        self.mmr_location = [3, 6]
        self.nan_location = [38, 39, 40]
        self.eplusminus_location = [1, 2, 4, 5]

        # SWA params
        hparams['scheduler_choice'] = 'swa' #'cycle' if 'scheduler_choice' not in hparams else hparams['scheduler_choice']
        hparams['save_freq'] = 25 if 'save_freq' not in hparams else hparams['save_freq']
        hparams['eval_freq'] = 5 if 'eval_freq' not in hparams else hparams['eval_freq']
        hparams['momentum'] = 0.9 if 'momentum' not in hparams else hparams['momentum']
        hparams['weight_decay'] = 1e-4 if 'weight_decay' not in hparams else hparams['weight_decay']
        hparams['noisy_val'] = True if 'noisy_val' not in hparams else hparams['noisy_val']

        # self.hparams = hparams
        self.save_hyperparameters()
        self.steps = hparams['steps']
        self.batch_size = hparams['batch_size']
        self.lr = hparams['lr'] #init_lr
        self._dataloader = None
        self._val_dataloader = None
        self.random_sample = False if 'random_sample' not in hparams else hparams['random_sample']
        self.train_len = 78660
        self.test_len = 8740
        self._summary_kl = 0.0
        self.include_mmr = hparams['include_mmr']
        self.include_nan = hparams['include_nan']
        self.include_eplusminus = True if 'include_eplusminus' not in hparams else hparams['include_eplusminus']
        self.train_all = False if 'train_all' not in hparams else hparams['train_all']

        self._cur_summary = None

        self.ssX = None
        self.ssy = None

    def augment(self, x):
        # This randomly samples times.
        samples = np.random.randint(self.hparams['samp'], x.shape[1]+1)
        x = x[:, np.random.randint(0, x.shape[1], size=samples)]
        return x

    def set_flag(self, flag_name, value):
        setattr(self, flag_name, value)
        for m in self.children():
            if hasattr(m, 'set_flag'):
                m.set_flag(flag_name, value)

    def compute_summary_stats(self, x):
        x = self.feature_nn(x)
        sample_mu = torch.mean(x, dim=1)
        sample_var = torch.std(x, dim=1)**2
        n = x.shape[1]
        
        std_in_mu = torch.sqrt(sample_var/n)
        std_in_var = torch.sqrt(2*sample_var**2/(n-1))
        
        # Take a "sample" of the average/variance of the learned features
        mu_sample =  torch.randn_like(sample_mu) *std_in_mu  + sample_mu
        var_sample = torch.randn_like(sample_var)*std_in_var + sample_var

        # Get to same unit
        std_sample = torch.sqrt(torch.abs(var_sample) + EPSILON)
        #clatent = torch.cat((mu_sample, var_sample), dim=1)
        clatent = torch.cat((mu_sample, std_sample), dim=1)
        self.latents = x

        return clatent

    def predict_instability(self, summary_stats):
        testy = self.regress_nn(summary_stats)
        # Outputs mu, std
        mu = soft_clamp(testy[:, [0]], 4.0, 12.0)
        std = soft_clamp(testy[:, [1]], self.lowest, 6.0)
        return mu, std

    def add_input_noise(self, x):
        noise = torch.randn_like(x, device=self.device) * torch.exp(self.input_noise_logvar[None, None, :]/2)
        return x + noise

    def add_summary_noise(self, summary_stats):
        noise = torch.randn_like(summary_stats, device=self.device) * torch.exp(self.summary_noise_logvar[None, :]/2)
        return summary_stats + noise

    def zero_megno(self, x):
        with torch.no_grad():
            mask = torch.zeros_like(x)
            mask[..., self.megno_location] = x[..., self.megno_location].clone()
            x = x - mask
        return x

    def zero_mmr(self, x):
        with torch.no_grad():
            mask = torch.zeros_like(x)
            mask[..., self.mmr_location] = x[..., self.mmr_location].clone()
            x = x - mask
        return x

    def zero_nan(self, x):
        with torch.no_grad():
            mask = torch.zeros_like(x)
            mask[..., self.nan_location] = x[..., self.nan_location].clone()
            x = x - mask
        return x

    def zero_eplusminus(self, x):
        with torch.no_grad():
            mask = torch.zeros_like(x)
            mask[..., self.eplusminus_location] = x[..., self.eplusminus_location].clone()
            x = x - mask
        return x

    def summarize_megno(self, x):
        megno_avg = torch.mean(x[:, :, [self.megno_location]], 1)
        megno_std = torch.std(x[:, :, [self.megno_location]], 1)

        return torch.cat([megno_avg, megno_std], dim=1)

    def forward(self, x, noisy_val=True):
        if self.fix_megno or self.fix_megno2:
            if self.fix_megno:
                megno_avg_std = self.summarize_megno(x)
            #(batch, 2)
            x = self.zero_megno(x)

        if not self.include_mmr:
            x = self.zero_mmr(x)

        if not self.include_nan:
            x = self.zero_nan(x)

        if not self.include_eplusminus:
            x = self.zero_eplusminus(x)

        if self.random_sample:
            x = self.augment(x)
        #x is (batch, time, feature)
        if noisy_val:
            x = self.add_input_noise(x)
        
        summary_stats = self.compute_summary_stats(x)
        if self.fix_megno:
            summary_stats = torch.cat([summary_stats, megno_avg_std], dim=1)

        self._cur_summary = summary_stats

        #summary is (batch, feature)
        self._summary_kl = (1/2) * (
                summary_stats**2
                + torch.exp(self.summary_noise_logvar)[None, :]
                - self.summary_noise_logvar[None, :]
                - 1
            )

        if noisy_val:
            summary_stats = self.add_summary_noise(summary_stats)

        mu, std = self.predict_instability(summary_stats)
        #Each is (batch,)

        return torch.cat((mu, std), dim=1)

    def sample(self, x, samples=10):
        all_samp = []
        init_settings = [self.random_sample, self.device]
        self.cpu()
        x = x.cpu()
        self.random_sample = False
        for _ in range(samples):
            out = self(x).detach().numpy()
            mu = out[:, 0]
            std = out[:, 1]
            all_samp.append(
                mu + np.random.randn(len(out))*std
            )
        self.random_sample = init_settings[0]
        self.to(init_settings[1])
        return np.average(all_samp, axis=0) 

    def _lossfnc(self, testy, y):
        mu = testy[:, [0]]
        std = testy[:, [1]]

        var = std**2
        t_greater_9 = y >= 9

        regression_loss = -(y - mu)**2/(2*var)
        regression_loss += -torch.log(std)
        regression_loss += -safe_log_erf(
                    (mu - 4)/(torch.sqrt(2*var))
                )
        classifier_loss = safe_log_erf(
                    (mu - 9)/(torch.sqrt(2*var))
            )

        safe_regression_loss = torch.where(
                ~torch.isfinite(regression_loss),
                -torch.ones_like(regression_loss)*100,
                regression_loss)
        safe_classifier_loss = torch.where(
                ~torch.isfinite(classifier_loss),
                -torch.ones_like(classifier_loss)*100,
                classifier_loss)

        total_loss = (
            safe_regression_loss * (~t_greater_9) +
            safe_classifier_loss * ( t_greater_9)
        )

        return -total_loss.sum(1)

    def lossfnc(self, x, y, samples=1, noisy_val=True):
        testy = self(x, noisy_val=noisy_val)
        n_samp = y.shape[0]
        loss = self._lossfnc(testy, y).sum()
        return loss

    def input_kl(self):
        return (1/2) * (
                torch.exp(self.input_noise_logvar)
                - self.input_noise_logvar
                - 1
            ).sum()

    def summary_kl(self):
        return self._summary_kl.sum()

    def training_step(self, batch, batch_idx):
        fraction = self.global_step / self.hparams['steps']
        beta_in = min([1, fraction/0.3]) * self.beta_in
        beta_out = min([1, fraction/0.3]) * self.beta_out

        X_sample, y_sample = batch
        loss = self.lossfnc(X_sample, y_sample, noisy_val=True)
        #cur_frac = len(X_sample) / self.train_len

        # Want to be important with total number of samples
        input_kl = self.input_kl() * beta_in * len(X_sample)
        summary_kl = self.summary_kl() * beta_out

        prior = input_kl + summary_kl

        total_loss = loss + prior

        tensorboard_logs = {'train_loss_no_reg': loss/len(X_sample), 'train_loss_with_reg': total_loss/len(X_sample), 'input_kl': input_kl/len(X_sample), 'summary_kl': summary_kl/len(X_sample)}

        return {'loss': total_loss, 'log': tensorboard_logs}

    def validation_step(self, batch, batch_idx):
        X_sample, y_sample = batch
        loss = self.lossfnc(X_sample, y_sample, noisy_val=self.hparams['noisy_val'])/self.test_len

        return {'val_loss': loss}

    def validation_epoch_end(self, outputs):

        avg_loss = torch.stack([x['val_loss'] for x in outputs]).sum()

        tensorboard_logs = {'val_loss_no_reg': avg_loss}

        return {'val_loss': avg_loss, 'log': tensorboard_logs}

    def configure_optimizers(self):
        opt1 = torch.optim.SGD(self.parameters(), lr=self.lr, momentum=self.hparams['momentum'], weight_decay=self.hparams['weight_decay'])

        assert self.hparams['scheduler_choice'] == 'swa'
        scheduler = CustomOneCycleLR(opt1, self.lr, int(0.9*self.steps), final_div_factor=1e4)
        interval = 'steps'
        name = 'swa_lr'

        sched1 = {
            'scheduler': scheduler,
            'name': name,
            'interval': interval
        }

        return [opt1], [sched1]

    def make_dataloaders(self, train=True, **extra_kwargs):
        kwargs = {
            **self.hparams,
            'model': self,
            **extra_kwargs,
            'train': train,
        }
        if 'ssX' in kwargs:
            dataloader, val_dataloader = get_data(**kwargs)
        else:
            dataloader, val_dataloader = get_data(ssX=self.ssX, **kwargs)

        labels = ['time', 'e+_near', 'e-_near', 'max_strength_mmr_near', 'e+_far', 'e-_far', 'max_strength_mmr_far', 'megno', 'a1', 'e1', 'i1', 'cos_Omega1', 'sin_Omega1', 'cos_pomega1', 'sin_pomega1', 'cos_theta1', 'sin_theta1', 'a2', 'e2', 'i2', 'cos_Omega2', 'sin_Omega2', 'cos_pomega2', 'sin_pomega2', 'cos_theta2', 'sin_theta2', 'a3', 'e3', 'i3', 'cos_Omega3', 'sin_Omega3', 'cos_pomega3', 'sin_pomega3', 'cos_theta3', 'sin_theta3', 'm1', 'm2', 'm3', 'nan_mmr_near', 'nan_mmr_far', 'nan_megno']
        for i in range(len(labels)):
            label = labels[i]
            if not ('cos' in label or
                'sin' in label or
                'nan_' in label or
                label == 'i1' or
                label == 'i2' or
                label == 'i3'):
                continue

            if not self.include_angles:
                print('Tossing', i, label)
                dataloader.dataset.tensors[0][..., i] = 0.0
                val_dataloader.dataset.tensors[0][..., i] = 0.0

        self._dataloader = dataloader
        self._val_dataloader = val_dataloader
        self.train_len = len(dataloader.dataset.tensors[0])
        self.test_len = len(val_dataloader.dataset.tensors[0])

    def train_dataloader(self):
        if self._dataloader is None:
            self.make_dataloaders()
        return self._dataloader

    def val_dataloader(self):
        if self._val_dataloader is None:
            self.make_dataloaders()
        return self._val_dataloader


class SWAGModel(VarModel):
    """Use .load_from_checkpoint(checkpoint_path) to initialize a SWAG model"""
    def init_params(self, swa_params):
        self.swa_params = swa_params
        self.swa_params['swa_lr'] = 0.001 if 'swa_lr' not in self.swa_params else self.swa_params['swa_lr']
        self.swa_params['swa_start'] = 1000 if 'swa_start' not in self.swa_params else self.swa_params['swa_start']
        self.swa_params['swa_recording_lr_factor'] = 0.5 if 'swa_recording_lr_factor' not in self.swa_params else self.swa_params['swa_recording_lr_factor']

        self.n_models = 0
        self.w_avg = None
        self.w2_avg = None
        self.pre_D = None
        self.K = 20 if 'K' not in self.swa_params else self.swa_params['K']
        self.c = 2 if 'c' not in self.swa_params else self.swa_params['c']
        self.swa_params['c'] = self.c
        self.swa_params['K'] = self.K

        return self

    def configure_optimizers(self):
        opt1 = torch.optim.SGD(self.parameters(), lr=self.swa_params['swa_lr'], momentum=self.hparams['momentum'], weight_decay=self.hparams['weight_decay'])
        scheduler = torch.optim.lr_scheduler.MultiStepLR(opt1, [self.swa_params['swa_start']], self.swa_params['swa_recording_lr_factor'])
        interval = 'steps'
        name = 'swa_record_lr'
        sched1 = {
            'scheduler': scheduler,
            'name': name,
            'interval': interval
        }

        return [opt1], [sched1]

    def training_step(self, batch, batch_idx):
        beta_in = self.beta_in
        beta_out = self.beta_out
        X_sample, y_sample = batch
        loss = self.lossfnc(X_sample, y_sample, noisy_val=True)
        input_kl = self.input_kl() * beta_in * len(X_sample)
        summary_kl = self.summary_kl() * beta_out
        prior = input_kl + summary_kl
        total_loss = loss + prior
        tensorboard_logs = {'train_loss_no_reg': loss/len(X_sample), 'train_loss_with_reg': total_loss/len(X_sample), 'input_kl': input_kl/len(X_sample), 'summary_kl': summary_kl/len(X_sample)}
        return {'loss': total_loss, 'log': tensorboard_logs}

    def flatten(self):
        """Convert state dict into a vector"""
        ps = self.state_dict()
        p_vec = None
        for key in ps.keys():
            p = ps[key]

            if p_vec is None:
                p_vec = p.reshape(-1)
            else:
                p_vec = torch.cat((p_vec, p.reshape(-1)))

        return p_vec

    def load(self, p_vec):
        """Load a vector into the state dict"""
        cur_state_dict = self.state_dict()
        new_state_dict = OrderedDict()
        i = 0
        for key in cur_state_dict.keys():
            old_p = cur_state_dict[key]
            size = old_p.numel()
            shape = old_p.shape
            new_p = p_vec[i:i+size].reshape(*shape)
            new_state_dict[key] = new_p
            i += size

        self.load_state_dict(new_state_dict)

    def aggregate_model(self):
        """Aggregate models for SWA/SWAG"""

        cur_w = self.flatten()
        cur_w2 = cur_w ** 2
        with torch.no_grad():
            if self.w_avg is None:
                self.w_avg = cur_w
                self.w2_avg = cur_w2
            else:
                self.w_avg = (self.w_avg * self.n_models + cur_w) / (self.n_models + 1)
                self.w2_avg = (self.w2_avg * self.n_models + cur_w2) / (self.n_models + 1)

            if self.pre_D is None:
                self.pre_D = cur_w.clone()[:, None]
            elif self.current_epoch % self.c == 0:
                #Record weights, measure discrepancy with average later
                self.pre_D = torch.cat((self.pre_D, cur_w[:, None]), dim=1)
                if self.pre_D.shape[1] > self.K:
                    self.pre_D = self.pre_D[:, 1:]
                    

        self.n_models += 1

    def validation_step(self, batch, batch_idx):
        X_sample, y_sample = batch
        loss = self.lossfnc(X_sample, y_sample, noisy_val=self.hparams['noisy_val'])/self.test_len

        if self.w_avg is None:
            swa_loss = loss
        else:
            tmp = self.flatten()
            self.load(self.w_avg)
            swa_loss = self.lossfnc(X_sample, y_sample, noisy_val=self.hparams['noisy_val'])/self.test_len
            self.load(tmp)

        return {'val_loss': loss, 'swa_loss': swa_loss}

    def validation_epoch_end(self, outputs):
        avg_loss = torch.stack([x['val_loss'] for x in outputs]).sum()
        swa_avg_loss = torch.stack([x['swa_loss'] for x in outputs]).sum()
        tensorboard_logs = {'val_loss_no_reg': avg_loss, 'swa_loss_no_reg': swa_avg_loss}
        #TODO: Check
        #fraction = self.global_step / self.hparams['steps']
        #if fraction > 0.5:
        if self.global_step > self.hparams['swa_start']:

            self.aggregate_model()

        # Record validation loss, and aggregated model loss
        return {'val_loss': avg_loss, 'log': tensorboard_logs}

    def sample_weights(self, scale=1):
        """Sample weights using SWAG:
        - w ~ N(avg_w, 1/2 * sigma + D . D^T/2(K-1))
            - This can be done with the following matrices:
                - z_1 ~ N(0, I_d); d the number of parameters
                - z_2 ~ N(0, I_K)
            - Then, compute:
            - w = avg_w + (1/sqrt(2)) * sigma^(1/2) . z_1 + D . z_2 / sqrt(2(K-1))
        """
        with torch.no_grad():
            avg_w = self.w_avg #[K]
            avg_w2 = self.w2_avg #[K]
            D = self.pre_D - avg_w[:, None]#[d, K]
            d = avg_w.shape[0]
            K = self.K
            z_1 = torch.randn((1, d), device=self.device)
            z_2 = torch.randn((K, 1), device=self.device)
            sigma = torch.abs(torch.diag(avg_w2 - avg_w**2))

            w = avg_w[None] + scale * (1.0/np.sqrt(2.0)) * z_1 @ sigma**0.5
            w += scale * (D @ z_2).T / np.sqrt(2*(K-1))
            w = w[0]

        self.load(w)

    def forward_swag(self, x, scale=0.5):
        """No augmentation happens here."""

        # Sample using SWAG using recorded model moments
        self.sample_weights(scale=scale)

        if self.fix_megno or self.fix_megno2:
            if self.fix_megno:
                megno_avg_std = self.summarize_megno(x)
            #(batch, 2)
            x = self.zero_megno(x)

        if not self.include_mmr:
            x = self.zero_mmr(x)

        if not self.include_nan:
            x = self.zero_nan(x)

        if not self.include_eplusminus:
            x = self.zero_eplusminus(x)

        summary_stats = self.compute_summary_stats(x)
        if self.fix_megno:
            summary_stats = torch.cat([summary_stats, megno_avg_std], dim=1)

        #summary is (batch, feature)
        self._summary_kl = (1/2) * (
                summary_stats**2
                + torch.exp(self.summary_noise_logvar)[None, :]
                - self.summary_noise_logvar[None, :]
                - 1
            )

        mu, std = self.predict_instability(summary_stats)
        #Each is (batch,)

        return torch.cat((mu, std), dim=1)

    def forward_swag_fast(self, x, scale=0.5):
        """No augmentation happens here."""

        # Sample using SWAG using recorded model moments
        self.sample_weights(scale=scale)

        if self.fix_megno or self.fix_megno2:
            if self.fix_megno:
                megno_avg_std = self.summarize_megno(x)
            #(batch, 2)
            x = self.zero_megno(x)

        if not self.include_mmr:
            x = self.zero_mmr(x)

        if not self.include_nan:
            x = self.zero_nan(x)

        if not self.include_eplusminus:
            x = self.zero_eplusminus(x)

        summary_stats = self.compute_summary_stats(x)
        if self.fix_megno:
            summary_stats = torch.cat([summary_stats, megno_avg_std], dim=1)

        #summary is (batch, feature)

        mu, std = self.predict_instability(summary_stats)
        #Each is (batch,)

        return torch.cat((mu, std), dim=1)


def save_swag(swag_model, path):
    save_items = {
        'hparams':swag_model.hparams,
        'swa_params': swag_model.swa_params,
        'w_avg': swag_model.w_avg.cpu(),
        'w2_avg': swag_model.w2_avg.cpu(),
        'pre_D': swag_model.pre_D.cpu()
    }

    torch.save(save_items, path)
    
def load_swag(path):
    save_items = torch.load(path)
    swag_model = (
        SWAGModel(save_items['hparams'])
        .init_params(save_items['swa_params'])
    )
    swag_model.w_avg = save_items['w_avg']
    swag_model.w2_avg = save_items['w2_avg']
    swag_model.pre_D = save_items['pre_D']
    
    return swag_model



import numpy as np
import sys
from subprocess import call
import time
import pandas as pd

Nruns = int(sys.argv[1])

try:
    df = pd.read_csv("../data/random/random.csv", index_col=0)
except:
    columns=['runstring']
    df = pd.DataFrame(columns=columns)

lastseed = df.shape[0]

for i in range(lastseed, lastseed+Nruns):
    runstring = "{0:0=7d}.bin".format(i)
    df.loc[i] = [runstring]
    with open("sunnyvale.sh", "w") as of:
        of.write("#!/bin/bash -l\n")
        of.write("#PBS -l nodes=1:ppn=1\n")
        of.write("#PBS -q workq\n")
        of.write("#PBS -r n\n")
        of.write("#PBS -l walltime=48:00:00\n")
        of.write("#PBS -N stab{0}\n".format(i))
        of.write("# EVERYTHING ABOVE THIS COMMENT IS NECESSARY, SHOULD ONLY CHANGE nodes,ppn,walltime and my_job_name VALUES\n")
        of.write("cd $PBS_O_WORKDIR\n")
        of.write("module load gcc/5.4.0\n")
        of.write("source /mnt/raid-cita/dtamayo/stability/bin/activate\n")
        of.write("python runrandom.py {0} {1}\n".format(i, runstring))

    call("chmod u=rwx sunnyvale.sh", shell=True)
    call("qsub sunnyvale.sh", shell=True)
        
df.to_csv("../data/random/random.csv", encoding='ascii')
